# ============================================================================
# ============================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Python implementations of preprocessing techniques.
# Contributors:
# ============================================================================

"""
Module of correction methods in the preprocessing stage:
- Flat-field correction.
- Distortion correction.
- MTF deconvolution.
- Tilted sinogram generation.
- Tilted 1D intensity-profile generation.
- Beam hardering correction.
"""

import numpy as np
import numpy.fft as fft
from scipy.ndimage import map_coordinates
import algotom.prep.removal as remo
import algotom.prep.filtering as filt
import algotom.util.utility as util


def flat_field_correction(proj, flat, dark, ratio=1.0, use_dark=True,
                          **options):
    """
    Do flat-field correction with options to remove zinger arifacts and/or
    stripe artifacts.

    Parameters
    ----------
    proj : array_like
        3D or 2D array. Projection images or a sinogram image.
    flat : array_like
        2D or 1D array. Flat-field image or a single row of it.
    dark : array_like
        2D or 1D array. Dark-field image or a single row of it.
    ratio : float
        Ratio between exposure time used for recording projections
        and exposure time used for recording flat field.
    use_dark : bool
        Subtracting dark field if True. May no need in some cases.
    options : dict, optional
        Apply a zinger removal method and/or ring removal methods.
        E.g option1={"method": "dezinger", "para1": 0.001, "para2": 1},
        option2={"method": "remove_stripe_based_sorting",
                "para1": 15, "para2": 1}

    Returns
    -------
    array_like
        3D or 2D array. Corrected projections or corrected sinograms.
    """
    flat = ratio * flat
    if use_dark:
        flatdark = flat - dark
        try:
            proj_corr = (np.float32(proj) - dark) / flatdark
        except ZeroDivisionError:
            nmean = np.mean(flatdark)
            if nmean != 0.0:
                flatdark[flatdark == 0.0] = nmean
            else:
                flatdark[flatdark == 0.0] = 1
            proj_corr = (np.float32(proj) - dark) / flatdark
    else:
        try:
            proj_corr = np.float32(proj) / flat
        except ZeroDivisionError:
            nmean = np.mean(flat)
            if nmean != 0.0:
                flat[flat == 0.0] = nmean
            else:
                flat[flat == 0.0] = 1
            proj_corr = np.float32(proj) / flat
    if len(options) != 0:
        for opt_name in options:
            opt = options[opt_name]
            if opt is not None:
                if 'method' in opt.keys():
                    method = opt['method']
                    list_para = tuple(opt.values())[1:]
                    if proj_corr.ndim == 2:
                        if method in dir(remo):
                            proj_corr = getattr(remo, method)(proj_corr, *list_para)
                        elif method in dir(filt):
                            proj_corr = getattr(filt, method)(proj_corr, *list_para)
                        else:
                            raise ValueError("Can't find the method: '{}' in the"
                                             " namespace".format(method))
                    else:
                        for i in np.arange(proj_corr.shape[1]):
                            if method in dir(remo):
                                proj_corr[:, i, :] = getattr(remo, method)(
                                    proj_corr[:, i, :], *list_para)
                            elif method in dir(filt):
                                proj_corr[:, i, :] = getattr(filt, method)(
                                    proj_corr[:, i, :], *list_para)
                            else:
                                raise ValueError("Can't find the method: '{}' in "
                                                 "the namespace".format(method))
                else:
                    raise ValueError("Incorrect option: {}".format(opt))
    return proj_corr


def unwarp_projection(proj, xcenter, ycenter, list_fact):
    """
    Apply distortion correction to a projection image using the polynomial
    backward model (Ref. [1]).

    Parameters
    ----------
    proj : array_like
        2D array. Projection image.
    xcenter : float
        Center of distortion in x-direction.
    ycenter : float
        Center of distortion in y-direction.
    list_fact : list of float
        Polynomial coefficients of the backward model.

    Returns
    -------
    array_like
        2D array. Distortion corrected.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.23.032859

    """
    (height, width) = proj.shape
    xu_list = np.arange(width) - xcenter
    yu_list = np.arange(height) - ycenter
    xu_mat, yu_mat = np.meshgrid(xu_list, yu_list)
    ru_mat = np.sqrt(xu_mat ** 2 + yu_mat ** 2)
    fact_mat = np.sum(np.asarray(
        [factor * ru_mat ** i for i, factor in enumerate(list_fact)]), axis=0)
    xd_mat = np.float32(np.clip(xcenter + fact_mat * xu_mat, 0, width - 1))
    yd_mat = np.float32(np.clip(ycenter + fact_mat * yu_mat, 0, height - 1))
    indices = np.reshape(yd_mat, (-1, 1)), np.reshape(xd_mat, (-1, 1))
    proj = map_coordinates(proj, indices, order=1, mode='reflect')
    return proj.reshape((height, width))


def unwarp_sinogram(data, index, xcenter, ycenter, list_fact, **option):
    """
    Unwarp sinogram [:,index.:] of a 3D tomographic dataset or
    a hdf/nxs object.

    Parameters
    ----------
    data : array_like or hdf object
        3D array.
    index : int
        Index of the sinogram.
    xcenter : float
        Center of distortion in x-direction.
    ycenter : float
        Center of distortion in y-direction.
    list_fact : list of float
        Polynomial coefficients of the backward model.
    option : list or tuple of int
        To extract subset data along axis 0 from a hdf object. E.g option =
        (start, stop, step)

    Returns
    -------
    array_like
        2D array. Distortion-corrected sinogram.
    """
    if data.ndim != 3:
        raise ValueError("Input must be a 3D data !!!")
    (depth, height, width) = data.shape
    if len(option) != 0:
        opt = option[list(option.keys())[0]]
        start, stop, step = opt
    else:
        start, stop, step = 0, depth, 1
    list_idx = range(start, stop, step)
    depth = len(list_idx)
    xu_list = np.arange(0, width) - xcenter
    yu = index - ycenter
    ru_list = np.sqrt(xu_list ** 2 + yu ** 2)
    flist = np.sum(np.asarray(
        [factor * ru_list ** i for i, factor in enumerate(list_fact)]), axis=0)
    xd_list = np.clip(xcenter + flist * xu_list, 0, width - 1)
    yd_list = np.clip(ycenter + flist * yu, 0, height - 1)
    yd_min = np.int16(np.floor(np.amin(yd_list)))
    yd_max = np.int16(np.ceil(np.amax(yd_list))) + 1
    yd_list = yd_list - yd_min
    indices = yd_list, xd_list
    sinogram = np.zeros((depth, width), dtype=np.float32)
    for i, idx in enumerate(list_idx):
        sinogram[i] = map_coordinates(
            data[idx, yd_min:yd_max, :], indices, order=1, mode='reflect')
    return sinogram


def unwarp_sinogram_chunk(data, start_index, stop_index, xcenter, ycenter,
                          list_fact, **option):
    """
    Unwarp chunk of sinograms [:, start_index: stop_index, :]
    of a 3D tomographic dataset or a hdf/nxs object.

    Parameters
    ----------
    data : array_like or hdf object
        3D array.
    start_index : int
        Starting index of sinograms.
    stop_index : int
        Stopping index of sinograms.
    xcenter : float
        Center of distortion in x-direction.
    ycenter : float
        Center of distortion in y-direction.
    list_fact : list of float
        Polynomial coefficients of the backward model.
    option : list or tuple of int
        To extract subset data along axis 0 from a hdf object. E.g option =
        [start, stop, step]

    Returns
    -------
    array_like
        3D array. Distortion corrected.
    """
    if data.ndim != 3:
        raise ValueError("Input must be a 3D data !!!")
    (depth, height, width) = data.shape
    if len(option) != 0:
        opt = option[list(option.keys())[0]]
        start, stop, step = opt
    else:
        start, stop, step = 0, depth, 1
    list_idx = range(start, stop, step)
    if stop_index == -1:
        stop_index = height
    xu_list = np.arange(0, width) - xcenter
    yu1 = start_index - ycenter
    ru_list = np.sqrt(xu_list ** 2 + yu1 ** 2)
    flist = np.sum(np.asarray(
        [factor * ru_list ** i for i, factor in enumerate(list_fact)]), axis=0)
    yd_list1 = np.clip(ycenter + flist * yu1, 0, height - 1)
    yu2 = stop_index - ycenter
    ru_list = np.sqrt(xu_list ** 2 + yu2 ** 2)
    flist = np.sum(np.asarray(
        [factor * ru_list ** i for i, factor in enumerate(list_fact)]), axis=0)
    yd_list2 = np.clip(ycenter + flist * yu2, 0, height - 1)
    yd_min = np.int16(np.floor(np.amin(yd_list1)))
    yd_max = np.int16(np.ceil(np.amax(yd_list2)))
    yu_list = np.arange(start_index, stop_index) - ycenter
    xu_mat, yu_mat = np.meshgrid(xu_list, yu_list)
    ru_mat = np.sqrt(xu_mat ** 2 + yu_mat ** 2)
    fact_mat = np.sum(np.asarray(
        [factor * ru_mat ** i for i, factor in enumerate(list_fact)]), axis=0)
    xd_mat = np.float32(np.clip(xcenter + fact_mat * xu_mat, 0, width - 1))
    yd_mat = np.float32(
        np.clip(ycenter + fact_mat * yu_mat, 0, height - 1)) - yd_min
    sino_chunk = np.asarray(
        [util.mapping(data[i, yd_min:yd_max, :], xd_mat, yd_mat) for i in
         list_idx])
    return sino_chunk


def mtf_deconvolution(mat, window, pad):
    """
    Deconvolve an projection image using division in the Fourier domain.
    Window can be determined using the approach in Ref. [1].

    Parameters
    ----------
    mat : array_like
        2D array. Projection image.
    window : array_like
        2D array. MTF function.
    pad : int
        Padding width to reduce the side effects of the Fourier transform.

    Returns
    -------
    array_like
        2D array. Deconvolved image.

    References
    ----------
    .. [1] https://doi.org/10.1117/12.2530324
    """
    (height1, width1) = mat.shape
    (height2, width2) = window.shape
    if (height1 > height2) or (width1 > width2):
        raise ValueError(
            "The sizes of the image are larger than the sizes of the window")
    else:
        pad_row = height2 - height1
        pad_col = width2 - width1
        mat = np.pad(mat, ((0, pad_row), (0, pad_col)), mode="edge")
    mat_pad = np.pad(mat, pad, mode="reflect")
    win_pad = np.pad(window, pad, mode="constant", constant_values=1.0)
    mat_dec = np.real(fft.ifft2(fft.fft2(mat_pad) / fft.ifftshift(win_pad)))
    mat_dec = mat_dec[pad: height2 + pad, pad: width2 + pad]
    return mat_dec[:height1, :width1]


def generate_tilted_sinogram(data, index, angle, **option):
    """
    Generate a tilted sinogram of a 3D tomographic dataset or a hdf/nxs object.

    Parameters
    ----------
    data : array_like or hdf object
        3D array.
    index : int
        Index of the sinogram.
    angle : float
        Tilted angle in degree.
    option : list or tuple of int
        To extract subset data along axis 0 from a hdf object. E.g option =
        (start, stop, step)

    Returns
    -------
    array_like
        2D array. Tilted sinogram.
    """
    if data.ndim != 3:
        raise ValueError("Input must be a 3D data !!!")
    (depth, height, width) = data.shape
    if len(option) != 0:
        opt = option[list(option.keys())[0]]
        start, stop, step = opt
    else:
        start, stop, step = 0, depth, 1
    list_idx = range(start, stop, step)
    depth = len(list_idx)
    x_cen = (width - 1.0) / 2
    y_cen = (height - 1.0) / 2
    x_list = np.arange(0, width) - x_cen
    y = index - y_cen
    angle = angle * np.pi / 180.0
    xt_list = x_list * np.cos(angle) - y * np.sin(angle)
    yt_list = x_list * np.sin(angle) + y * np.cos(angle)
    xt_list = np.clip(x_cen + xt_list, 0, width - 1)
    yt_list = np.clip(y_cen + yt_list, 0, height - 1)
    yt_min = np.int16(np.floor(np.amin(yt_list)))
    yt_max = np.int16(np.ceil(np.amax(yt_list))) + 1
    yt_list = yt_list - yt_min
    indices = yt_list, xt_list
    sinogram = np.zeros((depth, width), dtype=np.float32)
    for i, idx in enumerate(list_idx):
        sinogram[i] = map_coordinates(
            data[idx, yt_min:yt_max, :], indices, order=1, mode='reflect')
    return sinogram


def generate_tilted_sinogram_chunk(data, start_index, stop_index, angle,
                                   **option):
    """
    Generate a chunk of tilted sinograms of a 3D tomographic dataset or a
    hdf/nxs object.

    Parameters
    ----------
    data : array_like or hdf object
        3D array.
    start_index : int
        Starting index of sinograms.
    stop_index : int
        Stopping index of sinograms.
    angle : float
        Tilted angle in degree.
    option : list or tuple of int
        To extract subset data along axis 0 from a hdf object. E.g option =
        (start, stop, step)

    Returns
    -------
    array_like
        3D array. Chunk of tilted sinograms.
    """
    if data.ndim != 3:
        raise ValueError("Input must be a 3D data !!!")
    (depth, height, width) = data.shape
    if len(option) != 0:
        opt = option[list(option.keys())[0]]
        start, stop, step = opt
    else:
        start, stop, step = 0, depth, 1
    list_idx = range(start, stop, step)
    x_cen = (width - 1.0) / 2
    y_cen = (height - 1.0) / 2
    angle = angle * np.pi / 180.0
    x_list = np.arange(0, width) - x_cen
    y_list = np.arange(start_index, stop_index + 1) - y_cen
    x_mat, y_mat = np.meshgrid(x_list, y_list)
    x_mat1 = x_mat * np.cos(angle) - y_mat * np.sin(angle)
    y_mat1 = x_mat * np.sin(angle) + y_mat * np.cos(angle)
    x_mat1 = np.clip(x_cen + x_mat1, 0, width - 1)
    y_mat1 = np.clip(y_cen + y_mat1, 0, height - 1)
    y_min = np.int16(np.floor(np.amin(y_mat1)))
    y_max = np.int16(np.ceil(np.amax(y_mat1))) + 1
    y_mat1 = y_mat1 - y_min
    sino_chunk = np.asarray(
        [util.mapping(data[i, y_min:y_max, :], x_mat1, y_mat1) for i in
         list_idx])
    return sino_chunk


def generate_tilted_profile_line(mat, index, angle):
    """
    Generate a tilted horizontal intensity-profile of an image.

    Parameters
    ----------
    mat : array_like
        2D array.
    index : int
        Index of the line.
    angle : float
        Tilted angle in degree.

    Returns
    -------
    array_like
        1D array.
    """
    if mat.ndim != 2:
        raise ValueError("Input must be a 2D array !!!")
    (height, width) = mat.shape
    x_cen = (width - 1.0) / 2
    y_cen = (height - 1.0) / 2
    x_list = np.arange(0, width) - x_cen
    y = index - y_cen
    angle = angle * np.pi / 180.0
    xt_list = x_list * np.cos(angle) - y * np.sin(angle)
    yt_list = x_list * np.sin(angle) + y * np.cos(angle)
    xt_list = np.clip(x_cen + xt_list, 0, width - 1)
    yt_list = np.clip(y_cen + yt_list, 0, height - 1)
    yt_min = np.int16(np.floor(np.amin(yt_list)))
    yt_max = np.int16(np.ceil(np.amax(yt_list))) + 1
    yt_list = yt_list - yt_min
    indices = yt_list, xt_list
    profile_line = map_coordinates(mat[yt_min:yt_max, :], indices, order=1,
                                   mode='reflect')
    return profile_line


def generate_tilted_profile_chunk(mat, start_index, stop_index, angle):
    """
    Generate a chunk of tilted horizontal intensity-profiles of an image.

    Parameters
    ----------
    mat : array_like
        2D array.
    start_index : int
        Starting index of lines.
    stop_index : int
        Stopping index of lines.
    angle : float
        Tilted angle in degree.

    Returns
    -------
    array_like
        2D array.
    """
    if mat.ndim != 2:
        raise ValueError("Input must be a 2D data !!!")
    (height, width) = mat.shape
    x_cen = (width - 1.0) / 2
    y_cen = (height - 1.0) / 2
    angle = angle * np.pi / 180.0
    x_list = np.arange(0, width) - x_cen
    y_list = np.arange(start_index, stop_index + 1) - y_cen
    x_mat, y_mat = np.meshgrid(x_list, y_list)
    x_mat1 = x_mat * np.cos(angle) - y_mat * np.sin(angle)
    y_mat1 = x_mat * np.sin(angle) + y_mat * np.cos(angle)
    x_mat1 = np.clip(x_cen + x_mat1, 0, width - 1)
    y_mat1 = np.clip(y_cen + y_mat1, 0, height - 1)
    y_min = np.int16(np.floor(np.amin(y_mat1)))
    y_max = np.int16(np.ceil(np.amax(y_mat1))) + 1
    y_mat1 = y_mat1 - y_min
    profile_chunk = util.mapping(mat[y_min:y_max, :], x_mat1, y_mat1)
    return profile_chunk


def non_linear_function(intensity, q, n, opt=True):
    """
    Function used to define the response curve.

    Parameters
    ----------
    intensity : float
        Values stay in the range of [0; 1]
    q : float
        Positive number.
    n : float
        Positive number. Must larger than 1.
    opt : bool
        True: Curve more to values closer to 1.0.
        False: Curve more to values closer to 0.0
    Returns
    -------
    float
    """
    if opt:
        x = 1.0 - intensity
        num = np.log(1.0 - q * (1 - n))
        result = 1.0 - (np.log(1.0 - q * (x**n - n * x)) / num)
    else:
        x = intensity
        num = np.log(1.0 - q * (1 - n))
        result = (np.log(1.0 - q * (x**n - n * x)) / num)
    return result


def beam_hardening_correction(mat, q, n, opt=True):
    """
    Correct the grayscale values of a normalized image using a non-linear
    function.

    Parameters
    ----------
    mat : array_like
        Normalized projection image or sinogram image.
    q : float
        Positive number. Recommended range [0.005, 50].
    n : float
        Positive number. Must larger than 1.
    opt : bool
        True: Curve towards 0.0.
        False: Curve towards 1.0.

    Returns
    -------
    array_like
        Corrected image.
    """
    if np.max(mat) >= 2.0:
        raise ValueError("!!! Input image must be normalized, i.e. grayscales "
                         "are in the range of [0.0, 1.0]) !!!")
    if n < 2.0:
        raise ValueError("!!! n must be larger than or equal to 2 !!!")
    return np.asarray([non_linear_function(x, q, n, opt) for x in mat])



# ============================================================================
# ============================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Calibration methods
# Contributors:
# ============================================================================

"""
Module of calibration methods:
- Correcting the non-uniform background of an image.
- Binarizing an image.
- Calculating the distance between two point-like objects segmented from
  two images. Useful for determining pixel-size in helical scans.
"""

import numpy as np
import scipy.ndimage as ndi
import algotom.util.utility as util
import scipy.signal as sig


def normalize_background(mat, radius=51):
    """
    Correct a non-uniform background of an image using the median filter.

    Parameters
    ----------
    mat : array_like
        2D array.
    radius : int
        Size of the median filter.

    Returns
    -------
    array_like
        2D array. Corrected image.
    """
    mat_bck = ndi.median_filter(mat, radius, mode="reflect")
    mean_val = np.mean(mat_bck)
    try:
        mat_cor = mean_val * mat / mat_bck
    except ZeroDivisionError:
        mat_bck[mat_bck == 0.0] = mean_val
        mat_cor = mean_val * mat / mat_bck
    return mat_cor


def normalize_background_based_fft(mat, sigma=5, pad=None, mode="reflect"):
    """
    Correct a non-uniform background of an image using a Fourier Gaussian
    filter.

    Parameters
    ----------
    mat : array_like
        2D array.
    sigma : int
        Sigma of the Gaussian.
    pad : int
        Padding for the Fourier transform.
    mode : str, list of str, or tuple of str
        Padding method. One of options : 'reflect', 'edge', 'constant'. Full
        list is at:
        https://numpy.org/doc/stable/reference/generated/numpy.pad.html

    Returns
    -------
    array_like
        2D array. Corrected image.
    """
    (height, width) = mat.shape
    if height <= width:
        ratio = 1.0 * height / width
        sigmax = int(np.ceil(sigma / ratio))
        sigmay = sigma
    else:
        ratio = 1.0 * width / height
        sigmay = int(np.ceil(sigma / ratio))
        sigmax = sigma
    mat_bck = util.apply_gaussian_filter(mat, sigmax, sigmay,
                                         pad=pad, mode=mode)
    mean_val = np.mean(mat_bck)
    try:
        mat_cor = mean_val * mat / mat_bck
    except ZeroDivisionError:
        mat_bck[mat_bck == 0.0] = mean_val
        mat_cor = mean_val * mat / mat_bck
    return mat_cor


def invert_dot_contrast(mat):
    """
    Invert the contrast of a 2D binary array to make sure that a dot is white.

    Parameters
    ----------
    mat : array_like
        2D binary array.

    Returns
    -------
    array_like
        2D array.
    """
    (height, width) = mat.shape
    ratio = np.sum(mat) / (height * width)
    max_val = np.max(mat)
    if ratio > 0.5:
        mat = max_val - mat
    return mat


def calculate_threshold(mat, bgr="bright"):
    """
    Calculate threshold value based on Algorithm 4 in Ref. [1].

    Parameters
    ----------
    mat : array_like
        2D array.
    bgr : {"bright", "dark"}
        To indicate the brightness of the background against image features.

    Returns
    -------
    float
        Threshold value.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    size = max(mat.shape)
    list1 = np.sort(np.ndarray.flatten(mat))
    list1 = ndi.zoom(list1, (1.0 * size) / len(list1))
    list2 = sig.savgol_filter(list1, 2 * (len(list1) // 2) - 1, 3)
    if bgr == "bright":
        threshold = list2[0]
    else:
        threshold = list2[-1]
    return threshold


def binarize_image(mat, threshold=None, bgr="bright", norm=False, denoise=True,
                   invert=True):
    """
    Binarize an image.

    Parameters
    ----------
    mat : array_like
        2D array.
    threshold : float, optional
        Threshold value for binarization. Automatically calculated using
        Algorithm 4 in Ref. [1] if None.
    bgr : {"bright", "dark"}
        To indicate the brightness of the background against image features.
    norm : bool, optional
        Apply normalization if True.
    denoise : bool, optional
        Apply denoising if True.
    invert : bool, optional
        Invert the contrast if needed.

    Returns
    -------
    array_like
        2D binary array.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    if denoise is True:
        mat = ndi.median_filter(np.abs(mat), (3, 3))
    if norm is True:
        mat = normalize_background_based_fft(mat)
    if threshold is None:
        threshold = calculate_threshold(mat, bgr)
    else:
        num_min = np.min(mat)
        num_max = np.max(mat)
        if threshold < num_min or threshold > num_max:
            raise ValueError("Selected threshold value is out of the range of"
                             " [{0}, {1}]".format(num_min, num_max))
    mat = np.asarray(mat > threshold, dtype=np.float32)
    if invert is True:
        mat = invert_dot_contrast(mat)
    mat = np.int16(ndi.binary_fill_holes(mat))
    return mat


def get_dot_size(mat, size_opt="max"):
    """
    Get size of binary dots given the option.

    Parameters
    ----------
    mat : array_like
        2D binary array.
    size_opt : {"max", "min", "median", "mean"}
        Select options.

    Returns
    -------
    dot_size : float
        Size of the dot.
    """
    mat = np.int16(mat)
    mat_label, num_dots = ndi.label(mat)
    list_index = np.arange(1, num_dots + 1)
    list_sum = ndi.measurements.sum(mat, labels=mat_label, index=list_index)
    if size_opt == "median":
        dot_size = np.median(list_sum)
    elif size_opt == "mean":
        dot_size = np.mean(list_sum)
    elif size_opt == "min":
        dot_size = np.min(list_sum)
    else:
        dot_size = np.max(list_sum)
    return dot_size


def check_dot_size(mat, min_size, max_size):
    """
    Check if the size of a dot is in a range.

    Parameters
    ----------
    mat : array_like
        2D array.
    min_size : float
        Minimum size.
    max_size : float
        Maximum size.

    Returns
    -------
    bool
    """
    check = False
    dot_size = mat.sum()
    if (dot_size >= min_size) and (dot_size <= max_size):
        check = True
    return check


def select_dot_based_size(mat, dot_size, ratio=0.01):
    """
    Select dots having a certain size.

    Parameters
    ----------
    mat : array_like
        2D array.
    dot_size : float
        Size of the standard dot.
    ratio : float
        Used to calculate the acceptable range.
        [dot_size - ratio*dot_size; dot_size + ratio*dot_size]

    Returns
    -------
    array_like
        2D array. Selected dots.
    """
    mat = np.int16(mat)
    min_size = np.clip(np.int32(dot_size - ratio * dot_size), 1, None)
    max_size = np.clip(np.int32(dot_size + ratio * dot_size), 1, None)
    mat_label, _ = ndi.label(mat)
    list_dots = ndi.find_objects(mat_label)
    dots_selected = [dot for dot in list_dots
                     if check_dot_size(mat[dot], min_size, max_size)]
    mat1 = np.zeros_like(mat)
    for _, j in enumerate(dots_selected):
        mat1[j] = mat[j]
    return mat1


def calculate_distance(mat1, mat2, size_opt="max", threshold=None, bgr='bright',
                       norm=False, denoise=True, invert=True):
    """
    Calculate the distance between two point-like objects segmented from
    two images. Useful for measuring pixel-size in helical scans (Ref. [1]).

    Parameters
    ----------
    mat1 : array_like
        2D array.
    mat2 : array_like
        2D array.
    size_opt : {"max", "min", "median", "mean"}
        Options to select binary objects based on their size.
    threshold : float, optional
        Threshold value for binarization. Automatically calculated using
        Algorithm 4 in Ref. [2] if None.
    bgr : {"bright", "dark"}
        To indicate the brightness of the background against image features.
    norm : bool, optional
        Apply normalization if True.
    denoise : bool, optional
        Apply denoising if True.
    invert : bool, optional
        Invert the contrast if needed.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    .. [2] https://doi.org/10.1364/OE.26.028396
    """
    mat_bin1 = binarize_image(mat1, threshold=threshold, bgr=bgr, norm=norm,
                              denoise=denoise, invert=invert)
    dot_size1 = get_dot_size(mat_bin1, size_opt=size_opt)
    mat_bin1 = select_dot_based_size(mat_bin1, dot_size1)
    mat_bin2 = binarize_image(mat2, bgr=bgr, norm=norm, threshold=threshold,
                              denoise=denoise, invert=invert)
    dot_size2 = get_dot_size(mat_bin2, size_opt=size_opt)
    mat_bin2 = select_dot_based_size(mat_bin2, dot_size2)
    com1 = ndi.center_of_mass(mat_bin1)
    com2 = ndi.center_of_mass(mat_bin2)
    distance = np.sqrt((com1[0] - com2[0]) ** 2 + (com1[1] - com2[1]) ** 2)
    return distance



# ============================================================================
# ============================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Simulation methods
# Contributors:
# ============================================================================


"""
Module of simulation methods:
1- Methods for designing a customized 2D phantom.
2- Method for calculating a sinogram of a phantom based on the Fourier
   slice theorem.
3- Methods for adding artifacts to a simulated sinogram.
"""

import numpy as np
import scipy.signal.windows as win
import algotom.util.utility as util
import numpy.fft as fft


def make_elliptic_mask(size, center, length, angle):
    """
    Create an elliptic mask.

    Parameters
    -----------
    size : int
        Size of a square array.
    center : float or tuple of float
        Ellipse center.
    length : float or tuple of float
        Lengths of ellipse axes.
    angle : float
        Rotation angle (Degree) of the ellipse.

    Returns
    ------
    array_like
         Square array.
    """
    mask = np.zeros((size, size), dtype=np.float32)
    icenter = size // 2
    if isinstance(length, tuple):
        (x_len, y_len) = length
    else:
        x_len = y_len = length
    if isinstance(center, tuple):
        (x_cen, y_cen) = center
    else:
        x_cen = y_cen = center
    angle = - angle * np.pi / 180.0
    x_list = np.arange(size) - icenter - x_cen
    y_list = - np.arange(size) + icenter - y_cen
    x_mat, y_mat = np.meshgrid(x_list, y_list)
    x_mat1 = (x_mat * np.cos(angle) - y_mat * np.sin(angle)) / (0.5 * x_len)
    y_mat1 = (x_mat * np.sin(angle) + y_mat * np.cos(angle)) / (0.5 * y_len)
    r_mat = np.sqrt(x_mat1 ** 2 + y_mat1 ** 2)
    mask_check = r_mat <= 1.01
    mask[mask_check] = 1.0
    return mask


def make_rectangular_mask(size, center, length, angle):
    """
    Create a rectangular mask.

    Parameters
    -----------
    size : int
        Size of a square array.
    center : float or tuple of float
        Center of the mask.
    length : float or tuple of float
        Lengths of the rectangular mask.
    angle : float
        Rotation angle (Degree) of the mask.

    Returns
    ------
    array_like
         Square array.
    """
    mask = np.zeros((size, size), dtype=np.float32)
    icenter = size // 2
    if isinstance(length, tuple):
        (x_len, y_len) = length
    else:
        x_len = y_len = length
    if isinstance(center, tuple):
        (x_cen, y_cen) = center
    else:
        x_cen = y_cen = center
    angle = - angle * np.pi / 180.0
    x_list = np.arange(size) - icenter - x_cen
    y_list = - np.arange(size) + icenter - y_cen
    x_mat, y_mat = np.meshgrid(x_list, y_list)
    x_mat1 = np.abs(
        (x_mat * np.cos(angle) - y_mat * np.sin(angle)) / (0.5 * x_len))
    y_mat1 = np.abs(
        (x_mat * np.sin(angle) + y_mat * np.cos(angle)) / (0.5 * y_len))
    mask_check = (x_mat1 <= 1.01) & (y_mat1 <= 1.01)
    mask[mask_check] = 1.0
    return mask


def make_triangular_mask(size, center, length, angle):
    """
    Create an isosceles triangle mask.

    Parameters
    -----------
    size : int
        Size of a square array.
    center : float or tuple of float
        Center of the mask.
    length : float or tuple of float
        Lengths of the mask.
    angle : float
        Rotation angle (Degree) of the mask.

    Returns
    ------
    array_like
         Square array.
    """
    mask = make_rectangular_mask(size, center, length, angle)
    if isinstance(length, tuple):
        (x_len, y_len) = length
    else:
        x_len = y_len = length
    if isinstance(center, tuple):
        (x_cen, y_cen) = center
    else:
        x_cen = y_cen = center
    angle = np.deg2rad(angle)
    x_len1 = np.sqrt(x_len ** 2 + (0.5 * y_len) ** 2)
    angle1 = np.arctan2(0.5 * y_len, x_len)
    y_len1 = 2 * np.sin(angle1) * x_len
    x_off = - 0.5 * y_len1 * np.sin(angle1)
    y_off = 0.5 * y_len1 * np.cos(angle1) + 0.5 * x_len1 * np.sin(angle1)
    x_off1 = x_off * np.cos(angle) - y_off * np.sin(angle)
    y_off1 = x_off * np.sin(angle) + y_off * np.cos(angle)
    x_cen1 = x_cen + x_off1 + np.sign(x_off1) * 0.5
    y_cen1 = y_cen + y_off1 + np.sign(y_off1) * 0.5
    mask1 = make_rectangular_mask(size, (x_cen1, y_cen1), (x_len1, y_len1),
                                  np.rad2deg(angle + angle1))
    y_off = -y_off
    x_off1 = x_off * np.cos(angle) - y_off * np.sin(angle)
    y_off1 = x_off * np.sin(angle) + y_off * np.cos(angle)
    x_cen1 = x_cen + x_off1 + np.sign(x_off1) * 0.5
    y_cen1 = y_cen + y_off1 + np.sign(y_off1) * 0.5
    mask2 = make_rectangular_mask(size, (x_cen1, y_cen1), (x_len1, y_len1),
                                  np.rad2deg(angle - angle1))
    mask = np.clip(mask - mask1, 0.0, None)
    mask = np.clip(mask - mask2, 0.0, None)
    return mask


def make_line_target(size):
    """
    Create line patterns for testing the resolution of a reconstructed image.

    Parameters
    -----------
    size : int
        Size of a square array.

    Returns
    ------
    array_like
         Square array.
    """
    mask = np.zeros((size, size), dtype=np.float32)
    y_cen = 0
    line_hei = size // 16
    gap = 6
    check = True
    line_wid = 0
    start = line_hei // 2 + 1
    while check:
        line_wid = line_wid + 1
        stop = start + 3 * 2 * line_wid
        if stop > size // 2 - gap:
            check = False
        else:
            if line_wid % 2 == 1:
                for x_cen in np.arange(start, stop, 2 * line_wid):
                    mask += make_rectangular_mask(size, (x_cen, y_cen),
                                                  (line_wid, line_hei), 0.0)
            else:
                for x_cen in np.arange(start, stop, 2 * line_wid):
                    mask += make_rectangular_mask(size, (x_cen, y_cen),
                                                  (line_wid - 1, line_hei), 0.0)
                for x_cen in np.arange(start + line_wid // 2, stop,
                                       2 * line_wid):
                    mask += make_rectangular_mask(size, (x_cen, y_cen),
                                                  (1, line_hei), 0.0)
        start = stop + gap
    start = - line_hei // 2 - line_wid // 2
    line_wid = line_wid - 1
    while line_wid > 0:
        stop = start - 3 * 2 * line_wid
        if line_wid % 2 == 1:
            for x_cen in np.arange(start, stop, -2 * line_wid):
                mask += make_rectangular_mask(size, (x_cen, y_cen),
                                              (line_wid, line_hei), 0.0)
        else:
            for x_cen in np.arange(start, stop, -2 * line_wid):
                mask += make_rectangular_mask(size, (x_cen, y_cen),
                                              (line_wid - 1, line_hei), 0.0)
            for x_cen in np.arange(start - line_wid // 2, stop,
                                   -2 * line_wid):
                mask += make_rectangular_mask(size, (x_cen, y_cen),
                                              (1, line_hei), 0.0)
        start = stop - gap
        line_wid = line_wid - 1
    mask = mask + np.transpose(mask)
    list1 = mask[size // 2]
    list_pos = np.where(list1 == 1.0)[0]
    circle_mask = make_elliptic_mask(size, 0.0,
                                     list_pos[-1] - list_pos[0] + line_hei, 0.0)
    return circle_mask - mask


def make_face_phantom(size):
    """
    Create a face phantom for testing reconstruction methods.

    Parameters
    -----------
    size : int
        Size of a square array.

    Returns
    ------
    array_like
         Square array.
    """
    half = size // 2
    mask = np.zeros((size, size), dtype=np.float32)
    face1 = make_elliptic_mask(size, 0.0, (size / 1.3, 0.98 * size), 0.0)
    face2 = -0.6 * make_elliptic_mask(size, (0.0, -0.01 * size),
                                      (0.94 * size / 1.3, 0.94 * size), 0.0)
    face = face1 + face2
    x_rat_eye = 0.3
    y_rat_eye = 0.4
    eye1 = 0.6 * make_elliptic_mask(size, (-x_rat_eye * half, y_rat_eye * half),
                                    (0.15 * size, 0.05 * size), 0.0)
    pupil1a = -0.8 * make_elliptic_mask(size,
                                        (-x_rat_eye * half, y_rat_eye * half),
                                        (0.048 * size, 0.048 * size), 0.0)
    pupil1b = -0.2 * make_elliptic_mask(size,
                                        (-x_rat_eye * half, y_rat_eye * half),
                                        (0.015 * size, 0.015 * size), 0.0)
    pupil1 = pupil1a + pupil1b
    eyebrow1a = -0.3 * make_rectangular_mask(size,
                                             (-x_rat_eye * half, 0.54 * half),
                                             (0.18 * size, 0.02 * size), -5.0)
    eyebrow1b = -0.3 * make_rectangular_mask(size, (-x_rat_eye * half,
                                                    0.54 * half - 0.01 * half),
                                             (0.18 * size, 0.015 * size), -10.0)
    eyebrow1 = np.clip(eyebrow1a + eyebrow1b, -0.3, 0.0)
    eye2 = 0.6 * make_elliptic_mask(size, (x_rat_eye * half, y_rat_eye * half),
                                    (0.15 * size, 0.05 * size), 0.0)
    pupil2a = -0.8 * make_elliptic_mask(size,
                                        (x_rat_eye * half, y_rat_eye * half),
                                        (0.048 * size, 0.048 * size), 0.0)
    pupil2b = -0.2 * make_elliptic_mask(size,
                                        (x_rat_eye * half, y_rat_eye * half),
                                        (0.015 * size, 0.015 * size), 0.0)
    pupil2 = pupil2a + pupil2b
    eyebrow2a = -0.3 * make_rectangular_mask(size,
                                             (x_rat_eye * half, 0.54 * half),
                                             (0.18 * size, 0.02 * size), 5.0)
    eyebrow2b = -0.3 * make_rectangular_mask(size, (x_rat_eye * half,
                                                    0.54 * half - 0.01 * half),
                                             (0.18 * size, 0.015 * size), 10.0)
    eyebrow2 = np.clip(eyebrow2a + eyebrow2b, -0.3, 0.0)
    eye = eye1 + eye2 + pupil1 + pupil2 + eyebrow1 + eyebrow2
    nose1 = 0.2 * make_rectangular_mask(size, (0, 0),
                                        (0.05 * size, 0.25 * size), 0.0)
    nose2 = 0.2 * make_rectangular_mask(size, (0 + 0.015 * size, 0),
                                        (0.04 * size, 0.25 * size), 9.0)
    nose3 = 0.2 * make_rectangular_mask(size, (0 - 0.015 * size, 0),
                                        (0.04 * size, 0.25 * size), -9.0)
    nose = np.clip(nose1 + nose2 + nose3, 0.0, 0.2)
    mouth1 = 0.2 * make_rectangular_mask(size, (0.0, -0.22 * size),
                                         (0.24 * size, 0.055 * size), 0.0)
    mouth2 = 0.2 * make_elliptic_mask(size, (0.0, -0.22 * size + 0.025 * size),
                                      (0.24 * size, 0.07 * size), 0.0)
    mouth = mouth1 + mouth2
    mouth[mouth < 0.3] = 0.0
    beard1 = -0.4 * make_rectangular_mask(size, (0.0, -0.32 * size),
                                          (0.005 * size, 0.1 * size), 0.0)
    beard2 = -0.4 * make_rectangular_mask(size, (-0.02 * size, -0.32 * size),
                                          (0.005 * size, 0.1 * size), -10.0)
    beard3 = -0.4 * make_rectangular_mask(size, (0.02 * size, -0.32 * size),
                                          (0.005 * size, 0.1 * size), 10.0)
    beard = beard1 + beard2 + beard3
    mask += face + eye + nose + mouth + beard
    return mask


def make_sinogram(mat, angles, pad_rate=0.5, pad_mode="edge"):
    """
    Create a sinogram (series of 1D projections) from a 2D image based on the
    Fourier slice theorem (Ref. [1]).

    Parameters
    ----------
    mat : array_like
        Square array.
    angles : array_like
        1D array. List of angles (in radian) for projecting.
    pad_rate : float
        To apply padding before the FFT. The padding width equals to
        (pad_rate * image_width).
    pad_mode : str
        Padding method. Full list can be found at numpy.pad documentation.

    References
    ----------
    .. [1] https://doi.org/10.1071/PH560198
    """
    (nrow0, ncol0) = mat.shape
    if nrow0 != ncol0:
        raise ValueError(
            "Width and height of the image are not the same")
    if np.max(np.abs(angles)) > np.pi:
        print("!!! Warning !!!\nMaking sure that the angles are converted to "
              "Radian and in the range of [0; Pi]")
    pad = int(pad_rate * nrow0)
    mat_pad = np.pad(mat, pad, mode=pad_mode)
    if mat_pad.shape[0] % 2 == 0:
        mat_pad = np.pad(mat_pad, ((0, 1), (0, 1)), mode="edge")
    (nrow, ncol) = mat_pad.shape
    xcenter = (ncol - 1.0) * 0.5
    ycenter = (nrow - 1.0) * 0.5
    r_max = np.floor(max(xcenter, ycenter))
    r_list = np.linspace(-r_max, r_max, ncol)
    theta_list = - np.asarray(angles)
    r_mat, theta_mat = np.meshgrid(r_list, theta_list)
    x_mat = np.float32(
        np.clip(xcenter + r_mat * np.cos(theta_mat), 0, ncol - 1))
    y_mat = np.float32(
        np.clip(ycenter + r_mat * np.sin(theta_mat), 0, nrow - 1))
    mat_fft = fft.fftshift(fft.fft2(fft.ifftshift(mat_pad)))
    mat_real = np.real(mat_fft)
    mat_imag = np.imag(mat_fft)
    sino_real = util.mapping(mat_real, x_mat, y_mat, order=5, mode="reflect")
    sino_imag = util.mapping(mat_imag, x_mat, y_mat, order=5, mode="reflect")
    sinogram = np.real(fft.fftshift(
        fft.ifft(fft.ifftshift(sino_real + 1j * sino_imag, axes=1)), axes=1))
    return sinogram[:, pad:ncol0 + pad]


def add_noise(mat, noise_ratio=0.1):
    """
    Add Gaussian noise to an image.

    Parameters
    ----------
    mat : array_like
        2D array
    noise_ratio : float
        Ratio between the noise level and the mean of the array.

    Returns
    -------
    array_like
    """
    num_mean = np.mean(mat[mat != 0.0])
    noise_mean = num_mean * noise_ratio
    noise = np.random.normal(noise_mean, noise_mean * 0.5, size=mat.shape)
    return mat + noise


def add_stripe_artifact(sinogram, size, position, strength_ratio=0.2,
                        stripe_type="partial"):
    """
    Add stripe artifacts to a sinogram.

    Parameters
    ----------
    sinogram: array_like
        2D array. Sinogram image.
    size : int
        Size of stripe artifact.
    position : int
        Position of the stripe.
    strength_ratio : float
        To define the strength of the artifact. The value is in the range of
        [0.0, 1.0].
    stripe_type : {"partial", "full", "dead", "fluctuating"}
        Type of stripe as classified in Ref. [1].

    Returns
    -------
    array_like

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.26.028396
    """
    sinogram = np.copy(sinogram)
    (nrow, ncol) = sinogram.shape
    position = np.clip(position, 0, ncol - size - 1)
    strength_ratio = np.clip(strength_ratio, 0.0, 1.0)
    stripe = sinogram[:, position: position + size]
    if stripe_type == "partial":
        stripe_sort, mat_idx = util.sort_forward(stripe, axis=0)
        pos = int((1.0 - strength_ratio) * nrow)
        list_ratio = np.ones(nrow, dtype=np.float32)
        list_ratio[pos:nrow] = np.linspace(1.0, 1.0 - strength_ratio,
                                           nrow - pos)
        mat_ratio = np.tile(list_ratio, (size, 1))
        stripe_sort = stripe_sort * np.transpose(mat_ratio)
        stripe_mod = util.sort_backward(stripe_sort, mat_idx, axis=0)
    elif stripe_type == "dead":
        stripe_mod = np.ones_like(stripe) * strength_ratio * np.max(sinogram)
    elif stripe_type == "fluctuating":
        std_dev = np.mean(sinogram[sinogram != 0.0]) * strength_ratio
        noise = np.random.normal(0.0, std_dev, size=stripe.shape)
        stripe_mod = stripe + noise
    else:
        list_ratio = (1 - strength_ratio) * np.ones(nrow, dtype=np.float32)
        mat_ratio = np.tile(list_ratio, (size, 1))
        stripe_mod = stripe * np.transpose(mat_ratio)
    sinogram[:, position: position + size] = stripe_mod
    return sinogram


def convert_to_Xray_image(sinogram, global_max=None):
    """
    Convert a simulated sinogram to an equivalent X-ray image.

    Parameters
    ----------
    sinogram : array_like
        2D array.
    global_max : float
        Maximum value used for normalizing array values to stay in the range
        of [0.0, 1.0].

    Returns
    -------
    array_like
    """
    if global_max is None:
        global_max = np.max(sinogram)
    sinogram = 1.0 * sinogram / global_max
    return np.exp(-sinogram)


def add_background_fluctuation(sinogram, strength_ratio=0.2):
    """
    Fluctuate the background of a sinogram image using a Gaussian profile beam.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    strength_ratio : float
        To define the strength of the variation. The value is in the range of
        [0.0, 1.0].

    Returns
    -------
    array_like
    """
    sinogram = np.copy(sinogram)
    (nrow, ncol) = sinogram.shape
    list_fact = 1.0 - np.random.rand(nrow) * strength_ratio
    list_shift = np.int16(
        (0.5 - np.random.rand(nrow)) * strength_ratio * ncol * 0.5)
    for i in range(nrow):
        sinogram[i] = sinogram[i] * np.roll(
            win.gaussian(ncol, 0.5 * list_fact[i] * ncol), list_shift[i])
    return sinogram



# ============================================================================
# ============================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Python implementations of postprocessing techniques.
# Contributors:
# ============================================================================

"""
Module of methods in the postprocessing stage:
- Get statistical information of reconstructed images or a dataset.
- Downsample 2D, 3D array, or a dataset.
- Rescale 2D, 3D array or a dataset to 8-bit or 16-bit data-type.
- Removing ring artifacts in a reconstructed image by transform back and forth
between the polar coordinates and the Cartesian coordinates.
"""

import os
import numpy as np
from scipy.ndimage import gaussian_filter
import algotom.util.utility as util
import algotom.io.loadersaver as losa
import algotom.prep.removal as remo


def get_statical_information(mat, percentile=(5, 95), denoise=False):
    """
    Get statical information of an image.

    Parameters
    ----------
    mat : array_like
        2D array. Projection image, sinogram image, or reconstructed image.
    percentile : tuple of floats
        Tuple of (min_percentile, max_percentile) to compute.
        Must be between 0 and 100 inclusive.
    denoise: bool, optional
        Enable/disable denoising before extracting statistical information.

    Returns
    -------
    gmin : float
        The minimum value of the data array.
    gmax : float
        The maximum value of the data array.
    min_percent : float
        The first computed percentile of the data array.
    max_percent : tuple of floats
        The last computed percentile of the data array.
    mean : float
        The mean of the data array.
    median : float
        The median of the data array.
    variance : float
        The variance of the data array.
    """
    if denoise is True:
        mat = gaussian_filter(mat, 2)
    gmin = np.min(mat)
    gmax = np.max(mat)
    min_percent = np.percentile(mat, percentile[0])
    max_percent = np.percentile(mat, percentile[-1])
    median = np.median(mat)
    mean = np.mean(mat)
    variance = np.var(mat)
    return gmin, gmax, min_percent, max_percent, mean, median, variance


def get_statical_information_dataset(input_, percentile=(5, 95), skip=5,
                                     denoise=False, key_path=None):
    """
    Get statical information of a dataset. This can be a folder of tif files,
    a hdf file, or a 3D array.

    Parameters
    ----------
    input_ : str, hdf file, or array_like
        It can be a folder path to tif files, a hdf file, or a 3D array.
    percentile : tuple of floats
        Tuple of (min_percentile, max_percentile) to compute.
        Must be between 0 and 100 inclusive.
    skip : int
        Skipping step of reading input.
    denoise: bool, optional
        Enable/disable denoising before extracting statistical information.
    key_path : str, optional
        Key path to the dataset if the input is the hdf file.

    Returns
    -------
    gmin : float
        The global minimum value of the data array.
    gmax : float
        The global maximum value of the data array.
    min_percent : float
        The global min of the first computed percentile of the data array.
    max_percent : tuple of floats
        The global min of the last computed percentile of the data array.
    mean : float
        The mean of the data array.
    median : float
        The median of the data array.
    variance : float
        The mean of the variance of the data array.
    """
    if isinstance(input_, str) and (os.path.splitext(input_)[-1] == ""):
        list_file = losa.find_file(input_ + "/*.tif*")
        depth = len(list_file)
        if depth == 0:
            raise ValueError("No tif files in the folder: {}".format(input_))
        list_stat = []
        for i in range(0, depth, skip):
            mat = losa.load_image(list_file[i])
            if denoise is True:
                mat = gaussian_filter(mat, 2)
            list_stat.append(get_statical_information(mat, percentile, denoise))
    else:
        if isinstance(input_, str):
            file_ext = os.path.splitext(input_)[-1]
            if not (file_ext == '.hdf' or file_ext == '.h5'
                    or file_ext == ".nxs"):
                raise ValueError(
                    "Can't open this type of file format {}".format(file_ext))
            if key_path is None:
                raise ValueError(
                    "Please provide the key path to the dataset!!!")
            input_ = losa.load_hdf(input_, key_path)
        depth = len(input_)
        list_stat = []
        for i in range(0, depth, skip):
            mat = input_[i]
            if denoise is True:
                mat = gaussian_filter(mat, 2)
            list_stat.append(get_statical_information(mat, percentile, denoise))
    list_stat = np.asarray(list_stat)
    gmin = np.min(list_stat[:, 0])
    gmax = np.max(list_stat[:, 1])
    min_percent = np.min(list_stat[:, 2])
    max_percent = np.max(list_stat[:, 3])
    median = np.median(list_stat[:, 4])
    mean = np.mean(list_stat[:, 5])
    variance = np.mean(list_stat[:, 6])
    return gmin, gmax, min_percent, max_percent, mean, median, variance


def downsample(mat, cell_size, method="mean"):
    """
    Downsample an image.

    Parameters
    ----------
    mat : array_like
        2D array.
    cell_size : int or tuple of int
        Window size along axes used for grouping pixels.
    method : {"mean", "median", "max", "min"}
        Downsampling method.

    Returns
    -------
    array_like
        Downsampled image.
    """
    if method == "median":
        dsp_method = np.median
    elif method == "max":
        dsp_method = np.max
    elif method == "min":
        dsp_method = np.amin
    else:
        dsp_method = np.mean
    (height, width) = mat.shape
    if isinstance(cell_size, int):
        cell_size = (cell_size, cell_size)
    height_dsp = height // cell_size[0]
    width_dsp = width // cell_size[1]
    mat = mat[:height_dsp * cell_size[0], :width_dsp * cell_size[1]]
    mat_dsp = mat.reshape(
        height_dsp, cell_size[0], width_dsp, cell_size[1])
    mat_dsp = dsp_method(dsp_method(mat_dsp, axis=-1), axis=1)
    return mat_dsp


def downsample_dataset(input_, output, cell_size, method="mean", key_path=None):
    """
    Downsample a dataset. This can be a folder of tif files, a hdf file,
    or a 3D array.

    Parameters
    ----------
    input_ : str, array_like
        It can be a folder path to tif files, a hdf file, or 3D array.
    output : str, None
        It can be a folder path, a hdf file path, or None (memory consuming).
    cell_size : int or tuple of int
        Window size along axes used for grouping pixels.
    method : {"mean", "median", "max", "min"}
        Downsampling method.
    key_path : str, optional
        Key path to the dataset if the input is the hdf file.

    Returns
    -------
    array_like or None
        If output is None, returning an 3D array.
    """
    if output is not None:
        file_base, file_ext = os.path.splitext(output)
        if file_ext is not "":
            file_base = os.path.dirname(output)
        if os.path.exists(file_base):
            raise ValueError("Folder exists!!! Please choose another path!!!")
    if method == "median":
        dsp_method = np.median
    elif method == "max":
        dsp_method = np.max
    elif method == "min":
        dsp_method = np.amin
    else:
        dsp_method = np.mean
    if isinstance(cell_size, int):
        cell_size = (cell_size, cell_size, cell_size)
    if isinstance(input_, str) and (os.path.splitext(input_)[-1] == ""):
        list_file = losa.find_file(input_ + "/*.tif*")
        depth = len(list_file)
        if depth == 0:
            raise ValueError("No tif files in the folder: {}".format(input_))
        (height, width) = np.shape(losa.load_image(list_file[0]))
        depth_dsp = depth // cell_size[0]
        height_dsp = height // cell_size[1]
        width_dsp = width // cell_size[2]
        num = 0
        if (depth_dsp != 0) and (height_dsp != 0) and (width_dsp != 0):
            if output is not None:
                file_base, file_ext = os.path.splitext(output)
                if file_ext is not "":
                    if not (file_ext == '.hdf' or file_ext == '.h5'
                            or file_ext == ".nxs"):
                        raise ValueError(
                            "File extension must be hdf, h5, or nxs")
                output = file_base + file_ext
                data_out = losa.open_hdf_stream(
                    output, (depth_dsp, height_dsp, width_dsp),
                    key_path="downsample/data", overwrite=False)
            data_dsp = []
            for i in range(0, depth, cell_size[0]):
                if (i + cell_size[0]) > depth:
                    break
                else:
                    mat = []
                    for j in range(i, i + cell_size[0]):
                        mat.append(losa.load_image(list_file[j]))
                    mat = np.asarray(mat)
                    mat = mat[:, :height_dsp * cell_size[1],
                              :width_dsp * cell_size[2]]
                    mat = mat.reshape(1, cell_size[0], height_dsp,
                                      cell_size[1], width_dsp, cell_size[2])
                    mat_dsp = dsp_method(
                        dsp_method(dsp_method(mat, axis=-1), axis=1), axis=2)
                    if output is None:
                        data_dsp.append(mat_dsp[0])
                    else:
                        if file_ext is "":
                            out_name = "0000" + str(num)
                            losa.save_image(
                                output + "/img_" + out_name[-5:] + ".tif",
                                mat_dsp[0])
                        else:
                            data_out[num] = mat_dsp[0]
                        num += 1
        else:
            raise ValueError("Incorrect cell size {}".format(cell_size))
    else:
        if isinstance(input_, str):
            file_ext = os.path.splitext(input_)[-1]
            if not (file_ext == '.hdf' or file_ext == '.h5'
                    or file_ext == ".nxs"):
                raise ValueError(
                    "Can't open this type of file format {}".format(file_ext))
            if key_path is None:
                raise ValueError(
                    "Please provide the key path to the dataset!!!")
            input_ = losa.load_hdf(input_, key_path)
        (depth, height, width) = input_.shape
        depth_dsp = depth // cell_size[0]
        height_dsp = height // cell_size[1]
        width_dsp = width // cell_size[2]
        if (depth_dsp != 0) and (height_dsp != 0) and (width_dsp != 0):
            if output is None:
                input_ = input_[:depth_dsp * cell_size[0],
                                :height_dsp * cell_size[1],
                                :width_dsp * cell_size[2]]
                input_ = input_.reshape(
                    depth_dsp, cell_size[0], height_dsp, cell_size[1],
                    width_dsp, cell_size[2])
                data_dsp = dsp_method(
                    dsp_method(dsp_method(input_, axis=-1), axis=1), axis=2)
            else:
                file_base, file_ext = os.path.splitext(output)
                if file_ext is not "":
                    if not (file_ext == '.hdf' or file_ext == '.h5'
                            or file_ext == ".nxs"):
                        raise ValueError(
                            "File extension must be hdf, h5, or nxs")
                    output = file_base + file_ext
                    data_out = losa.open_hdf_stream(
                        output, (depth_dsp, height_dsp, width_dsp),
                        key_path="downsample/data", overwrite=False)
                num = 0
                for i in range(0, depth, cell_size[0]):
                    if (i + cell_size[0]) > depth:
                        break
                    else:
                        mat = input_[i:i + cell_size[0],
                                     :height_dsp * cell_size[1],
                                     :width_dsp * cell_size[2]]
                        mat = mat.reshape(1, cell_size[0], height_dsp,
                                          cell_size[1], width_dsp, cell_size[2])
                        mat_dsp = dsp_method(dsp_method(
                            dsp_method(mat, axis=-1), axis=1), axis=2)
                        if file_ext is not "":
                            data_out[num] = mat_dsp[0]
                        else:
                            out_name = "0000" + str(num)
                            losa.save_image(
                                output + "/img_" + out_name[-5:] + ".tif",
                                mat_dsp[0])
                        num += 1
        else:
            raise ValueError("Incorrect cell size {}".format(cell_size))
    if output is None:
        return np.asarray(data_dsp)


def rescale(mat, nbit=16, minmax=None):
    """
    Rescale a 32-bit array to 16-bit/8-bit data.

    Parameters
    ----------
    mat : array_like
    nbit : {8,16}
        Rescaled data-type: 8-bit or 16-bit.
    minmax : tuple of float, or None
        Minimum and maximum values used for rescaling.

    Returns
    -------
    array_like
        Rescaled array.
    """
    if minmax is None:
        gmin, gmax = np.min(mat), np.max(mat)
    else:
        (gmin, gmax) = minmax
    mat = np.clip(mat, gmin, gmax)
    mat = (mat - gmin) / (gmax - gmin)
    if nbit == 8:
        mat = np.uint8(np.clip(mat * 255, 0, 255))
    else:
        mat = np.uint16(np.clip(mat * 65535, 0, 65535))
    return mat


def rescale_dataset(input_, output, nbit=16, minmax=None, skip=None,
                    key_path=None):
    """
    Rescale a dataset to 8-bit or 16-bit data-type. The dataset can be a
    folder of tif files, a hdf file, or a 3D array.

    Parameters
    ----------
    input_ : str, array_like
        It can be a folder path to tif files, a hdf file, or 3D array.
    output : str, None
        It can be a folder path, a hdf file path, or None (memory consuming).
    nbit : {8,16}
        Rescaled data-type: 8-bit or 16-bit.
    minmax : tuple of float, or None
        Minimum and maximum values used for rescaling. They are calculated if
        None is given.
    skip : int or None
        Skipping step of reading input used for getting statistical information.
    key_path : str, optional
        Key path to the dataset if the input is the hdf file.

    Returns
    -------
    array_like or None
        If output is None, returning an 3D array.
    """
    if output is not None:
        file_base, file_ext = os.path.splitext(output)
        if file_ext is not "":
            file_base = os.path.dirname(output)
        if os.path.exists(file_base):
            raise ValueError("Folder exists!!! Please choose another path!!!")
    if isinstance(input_, str) and (os.path.splitext(input_)[-1] == ""):
        list_file = losa.find_file(input_ + "/*.tif*")
        depth = len(list_file)
        if depth == 0:
            raise ValueError("No tif files in the folder: {}".format(input_))
        if minmax is None:
            if skip is None:
                skip = int(np.ceil(0.15 * depth))
            (gmin, gmax) = get_statical_information_dataset(input_, skip=skip)[
                           0:2]
        else:
            (gmin, gmax) = minmax
        if output is not None:
            file_base, file_ext = os.path.splitext(output)
            if file_ext is not "":
                if not (file_ext == '.hdf' or file_ext == '.h5'
                        or file_ext == ".nxs"):
                    raise ValueError("File extension must be hdf, h5, or nxs")
                output = file_base + file_ext
                (height, width) = np.shape(losa.load_image(list_file[0]))
                if nbit == 8:
                    data_type = "uint8"
                else:
                    data_type = "uint16"
                data_out = losa.open_hdf_stream(output, (depth, height, width),
                                                key_path="rescale/data",
                                                data_type=data_type,
                                                overwrite=False)
        data_res = []
        for i in range(0, depth):
            mat = rescale(
                losa.load_image(list_file[i]), nbit=nbit, minmax=(gmin, gmax))
            if output is None:
                data_res.append(mat)
            else:
                file_base, file_ext = os.path.splitext(output)
                if file_ext is "":
                    out_name = "0000" + str(i)
                    losa.save_image(output + "/img_" + out_name[-5:] + ".tif",
                                    mat)
                else:
                    data_out[i] = mat
    else:
        if isinstance(input_, str):
            file_ext = os.path.splitext(input_)[-1]
            if not (file_ext == '.hdf' or file_ext == '.h5'
                    or file_ext == ".nxs"):
                raise ValueError(
                    "Can't open this type of file format {}".format(file_ext))
            if key_path is None:
                raise ValueError(
                    "Please provide the key path to the dataset!!!")
            input_ = losa.load_hdf(input_, key_path)
        (depth, height, width) = input_.shape
        if minmax is None:
            if skip is None:
                skip = int(np.ceil(0.15 * depth))
            (gmin, gmax) = get_statical_information_dataset(input_, skip=skip,
                                                            key_path=key_path)[
                           0:2]
        else:
            (gmin, gmax) = minmax
        data_res = []
        if output is not None:
            file_base, file_ext = os.path.splitext(output)
            if file_ext is not "":
                if not (file_ext == '.hdf' or file_ext == '.h5'
                        or file_ext == ".nxs"):
                    raise ValueError("File extension must be hdf, h5, or nxs")
                output = file_base + file_ext
                if nbit == 8:
                    data_type = "uint8"
                else:
                    data_type = "uint16"
                data_out = losa.open_hdf_stream(
                    output, (depth, height, width), key_path="rescale/data",
                    data_type=data_type, overwrite=False)
        for i in range(0, depth):
            mat = rescale(input_[i], nbit=nbit, minmax=(gmin, gmax))
            if output is None:
                data_res.append(mat)
            else:
                file_base, file_ext = os.path.splitext(output)
                if file_ext is not "":
                    data_out[i] = mat
                else:
                    out_name = "0000" + str(i)
                    losa.save_image(output + "/img_" + out_name[-5:] + ".tif",
                                    mat)
    if output is None:
        return np.asarray(data_res)


def remove_ring_based_fft(mat, u=20, n=8, v=1, sort=False):
    """
    Remove ring artifacts in the reconstructed image by combining the polar
    transform and the fft-based method.

    Parameters
    ----------
    mat : array_like
        Square array. Reconstructed image
    u : int
        Cutoff frequency.
    n : int
        Filter order.
    v : int
        Number of rows (* 2) to be applied the filter.
    sort : bool, optional
        Apply sorting (Ref. [2]) if True.

    Returns
    -------
    array_like
        Ring-removed image.

    References
    ----------
    .. [1] https://doi.org/10.1063/1.1149043
    .. [2] https://doi.org/10.1364/OE.26.028396
    """
    (nrow, ncol) = mat.shape
    if nrow != ncol:
        raise ValueError(
            "Width and height of the reconstructed image are not the same")
    mask = util.make_circle_mask(ncol, 1.0)
    (x_mat, y_mat) = util.rectangular_from_polar(ncol, ncol, ncol, ncol)
    (r_mat, theta_mat) = util.polar_from_rectangular(ncol, ncol, ncol, ncol)
    polar_mat = util.mapping(mat, x_mat, y_mat)
    polar_mat = remo.remove_stripe_based_fft(polar_mat, u, n, v, sort=sort)
    mat_rec = util.mapping(polar_mat, r_mat, theta_mat)
    return mat_rec * mask


def remove_ring_based_wavelet_fft(mat, level=5, size=1, wavelet_name="db9",
                                  sort=False):
    """
    Remove ring artifacts in a reconstructed image by combining the polar
    transform and the wavelet-fft-based method (Ref. [1]).

    Parameters
    ----------
    mat : array_like
        Square array. Reconstructed image
    level : int
        Wavelet decomposition level.
    size : int
        Damping parameter. Larger is stronger.
    wavelet_name : str
        Name of a wavelet. Search pywavelets API for a full list.
    sort : bool, optional
        Apply sorting (Ref. [2]) if True.

    Returns
    -------
    array_like
        Ring-removed image.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.17.008567
    .. [2] https://doi.org/10.1364/OE.26.028396
    """
    (nrow, ncol) = mat.shape
    if nrow != ncol:
        raise ValueError(
            "Width and height of the reconstructed image are not the same")
    mask = util.make_circle_mask(ncol, 1.0)
    (x_mat, y_mat) = util.rectangular_from_polar(ncol, ncol, ncol, ncol)
    (r_mat, theta_mat) = util.polar_from_rectangular(ncol, ncol, ncol, ncol)
    polar_mat = util.mapping(mat, x_mat, y_mat)
    polar_mat = remo.remove_stripe_based_wavelet_fft(polar_mat, level, size,
                                                     wavelet_name, sort=sort)
    mat_rec = util.mapping(polar_mat, r_mat, theta_mat)
    return mat_rec * mask



# ============================================================================
# ============================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Python implementations of preprocessing techniques.
# Contributors:
# ============================================================================

"""
Module of filtering methods in the preprocessing stage:
- Fresnel filter (denoising or low-pass filter).
- Double-wedge filter.
"""

import numpy as np
import numpy.fft as fft
import algotom.prep.conversion as conv


def make_fresnel_window(height, width, ratio, dim):
    """
    Create a low pass window based on the Fresnel propagator.
    It is used to denoise a projection image (dim=2) or a
    sinogram image (dim=1).

    Parameters
    ----------
    height : int
        Image height
    width : int
        Image width
    ratio : float
        To define the shape of the window.
    dim : {1, 2}
        Use "1" if working on a sinogram image and "2" if working on
        a projection image.

    Returns
    -------
    array_like
        2D array.
    """
    ycenter = int(np.ceil((height - 1) * 0.5))
    xcenter = int(np.ceil((width - 1) * 0.5))
    if dim == 2:
        ulist = (1.0 * np.arange(0, width) - xcenter) / width
        vlist = (1.0 * np.arange(0, height) - ycenter) / height
        u, v = np.meshgrid(ulist, vlist)
        window = 1.0 + ratio * (u ** 2 + v ** 2)
    else:
        ulist = (1.0 * np.arange(0, width) - xcenter) / width
        win1d = 1.0 + ratio * ulist ** 2
        window = np.tile(win1d, (height, 1))
    return window


def fresnel_filter(mat, ratio, dim=1, window=None, pad=150, apply_log=True):
    """
    Apply a low-pass filter based on the Fresnel propagator to an image
    (Ref. [1]).

    Parameters
    ----------
    mat : array_like
        2D array. Projection image or sinogram image.
    ratio : float
        Define the shape of the window. Larger is more smoothing.
    dim : {1, 2}
        Use "1" if working on a sinogram image and "2" if working on
        a projection image.
    window : array_like, optional
        Window for deconvolution.
    pad : int
        Padding width.
    apply_log : bool, optional
        Apply the logarithm function to the sinogram before filtering.

    Returns
    -------
    array_like
        2D array. Filtered image.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    """
    if apply_log:
        mat = -np.log(mat)
    (nrow, ncol) = mat.shape
    if dim == 2:
        if window is None:
            window = make_fresnel_window(nrow, ncol, ratio, dim)
        mat_pad = np.pad(mat, pad, mode="edge")
        win_pad = np.pad(window, pad, mode="edge")
        mat_dec = fft.ifft2(fft.fft2(mat_pad) / fft.ifftshift(win_pad))
        mat_dec = np.real(mat_dec[pad:pad + nrow, pad:pad + ncol])
    else:
        if window is None:
            window = make_fresnel_window(nrow, ncol, ratio, dim)
        mat_pad = np.pad(mat, ((0, 0), (pad, pad)), mode='edge')
        win_pad = np.pad(window, ((0, 0), (pad, pad)), mode="edge")
        mat_fft = np.fft.fftshift(fft.fft(mat_pad), axes=1) / win_pad
        mat_dec = fft.ifft(np.fft.ifftshift(mat_fft, axes=1))
        mat_dec = np.real(mat_dec[:, pad:pad + ncol])
    if apply_log:
        mat_dec = np.exp(-mat_dec)
    return np.float32(mat_dec)


def make_double_wedge_mask(height, width, radius):
    """
    Generate a double-wedge binary mask using Eq. (3) in Ref. [1].
    Values outside the double-wedge region correspond to 0.0.

    Parameters
    ----------
    height : int
        Image height.
    width : int
        Image width.
    radius : int
        Radius of an object, in pixel unit.

    Returns
    -------
    array_like
        2D binary mask.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.22.019078
    """
    du = 1.0 / width
    dv = (height - 1.0) / (height * 2.0 * np.pi)
    ndrop = 1
    ycenter = np.int16(np.ceil((height - 1) / 2.0))
    xcenter = np.int16(np.ceil((width - 1) / 2.0))
    mask = np.ones((height, width), dtype=np.float32)
    for i in range(height):
        num = np.int16(np.ceil(((i - ycenter) * dv / radius) / du))
        (p1, p2) = np.int16(np.clip(
            np.sort((-num + xcenter, num + xcenter)), 0, width - 1))
        mask[i, p1:p2 + 1] = 0.0
    mask[ycenter - ndrop:ycenter + ndrop + 1, :] = 1.0
    return mask


def double_wedge_filter(sinogram, center=0, sino_type="180", iteration=5,
                        mask=None, ratio=1.0, pad=250):
    """
    Apply double-wedge filter to a sinogram image (Ref. [1]).

    Parameters
    ----------
    sinogram : array_like
        2D array. 180-degree sinogram or 360-degree sinogram.
    center : float, optional
        Center-of-rotation. No need for a 360-sinogram.
    sino_type : {"180", "360"}
        Sinogram type : 180-degree or 360-degree.
    iteration : int
        Number of iteration.
    mask : array_like, optional
        Double-wedge binary mask.
    ratio : float, optional
        Define the cut-off angle of the double-wedge filter.
    pad : int
        Padding width.

    Returns
    -------
    array_like
        2D array. Filtered sinogram.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    """
    if not (sino_type == "180" or sino_type == "360"):
        raise ValueError("!!! Use only one of two options: '180' or '360'!!!")
    if sino_type == "180":
        nrow0 = sinogram.shape[0]
        if center == 0:
            raise ValueError(
                "Please provide the location of the rotation axis")
        sinogram = conv.convert_sinogram_180_to_360(sinogram, center)
    (nrow, ncol) = sinogram.shape
    ncol_pad = ncol + 2 * pad
    if mask is None:
        mask = make_double_wedge_mask(nrow, ncol_pad, ratio * ncol / 2.0)
    else:
        if mask.shape != (nrow, ncol_pad):
            raise ValueError(
                "Shape of the left-right padded sinogram {0} and the mask "
                "{1} is not the same!!!".format((nrow, ncol_pad), mask.shape))
    sino_filt = np.copy(sinogram)
    for i in range(iteration):
        sino_filt = np.pad(sino_filt, ((0, 0), (pad, pad)), mode="edge")
        sino_filt = np.real(fft.ifft2(
            fft.ifftshift(fft.fftshift(fft.fft2(sino_filt)) * mask)))
        sino_filt = sino_filt[:, pad:ncol + pad]
    if sino_type == "180":
        sino_filt = sino_filt[:nrow0]
    return sino_filt



import pathlib
import setuptools

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setuptools.setup(
    name="algotom",
    version="1.0.2",
    author="Nghia Vo",
    author_email="nghia.vo@diamond.ac.uk",
    description="Data processing algorithms for tomography",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=["Parallel-beam Computed Tomography", "Image Processing",
              "Tomography", "X-ray Imaging"],
    url="https://github.com/algotom/algotom",
    license="Apache 2.0",
    platforms="Any",
    packages=setuptools.find_packages(include=["algotom","algotom.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Natural Language :: English",
        "Topic :: Scientific/Engineering :: Image Processing"
    ],
    install_requires=[
        "numpy>=1.15",
        "scipy",
        "numba>=0.50.1",
        "pywavelets",
        "pillow",
        "h5py",
        "joblib"
    ],
    python_requires='>=3.7',
)



# ============================================================================
# ============================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Python implementations of FFT-based reconstruction methods
# Contributors:
# ============================================================================

"""
Module of FFT-based reconstruction methods in the reconstruction stage:
- Filtered back-projection (FBP) method for GPU (using numba and cuda) and CPU.
- Direct Fourier inversion (DFI) method.
- Wrapper for Astra Toolbox reconstruction (optional)
- Wrapper for Tomopy-gridrec reconstruction (optional)
"""

import math
import numpy as np
from numba import cuda
from numba import jit
from scipy import signal
from scipy.ndimage import shift
import numpy.fft as fft
import algotom.util.utility as util


def make_smoothing_window(filter_name, width):
    """
    Make a 1d smoothing window.

    Parameters
    ----------
    filter_name : {"hann", "bartlett", "blackman", "hamming", "nuttall",\\
                   "parzen", "triang"}
        Window function used for filtering.
    width : int
        Width of the window.

    Returns
    -------
    array_like
        1D array.
    """
    if filter_name == 'hann':
        window = signal.windows.hann(width)
    elif filter_name == 'bartlett':
        window = signal.windows.bartlett(width)
    elif filter_name == 'blackman':
        window = signal.windows.blackman(width)
    elif filter_name == 'hamming':
        window = signal.windows.hamming(width)
    elif filter_name == 'nuttall':
        window = signal.windows.nuttall(width)
    elif filter_name == 'parzen':
        window = signal.windows.parzen(width)
    elif filter_name == 'triang':
        window = signal.windows.triang(width)
    else:
        window = np.ones(width)
    return window


def make_2d_ramp_window(height, width, filter_name=None):
    """
    Make the 2d ramp window (in the Fourier space) by repeating the 1d ramp
    window with the option of adding a smoothing window.

    Parameters
    ----------
    height : int
        Height of the window.
    width : int
        Width of the window.
    filter_name : {None, "hann", "bartlett", "blackman", "hamming", "nuttall",\\
                  "parzen", "triang"}
         Name of a smoothing window used.

    Returns
    -------
    complex ndarray
        2D array.
    """
    ramp_win = np.arange(0.0, width) - np.ceil((width - 1.0) / 2)
    ramp_win[ramp_win == 0.0] = 0.25
    ramp_win[ramp_win % 2 == 0.0] = 0.0
    for i in range(width):
        if ramp_win[i] % 2 == 1.0:
            ramp_win[i] = - 1.0 / (ramp_win[i] * np.pi) ** 2
    window = make_smoothing_window(filter_name, width)
    ramp_fourier = fft.fftshift(fft.fft(ramp_win)) * window
    ramp_fourier_2d = np.tile(ramp_fourier, (height, 1))
    return ramp_fourier_2d


def apply_ramp_filter(sinogram, ramp_win=None, filter_name=None, pad=None,
                      pad_mode="edge"):
    """
    Apply the ramp filter to a sinogram with the option of adding a smoothing
    filter.

    Parameters
    ----------
    sinogram : array_like
        2D rray. Sinogram image.
    ramp_win : complex ndarray or None
        Ramp window in the Fourier space.
    filter_name : {None, "hann", "bartlett", "blackman", "hamming", "nuttall",\\
                  "parzen", "triang"}
         Name of a smoothing window used.
    pad : int or None
        To apply padding before the FFT. The value is set to 10% of the image
        width if None is given.
    pad_mode : str
        Padding method. Full list can be found at numpy.pad documentation.

    Returns
    -------
    array_like
        Filtered sinogram.
    """
    (nrow, ncol) = sinogram.shape
    if pad is None:
        pad = int(0.1 * ncol)
    sino_pad = np.pad(sinogram, ((0, 0), (pad, pad)), mode=pad_mode)
    if (ramp_win is None) or (ramp_win.shape != sinogram.shape):
        ramp_win = make_2d_ramp_window(nrow, ncol + 2 * pad, filter_name)
    sino_fft = fft.fftshift(fft.fft(sino_pad), axes=1) * ramp_win
    sino_filtered = np.real(
        fft.ifftshift(fft.ifft(fft.ifftshift(sino_fft, axes=1)), axes=1))
    return np.ascontiguousarray(sino_filtered[:, pad:ncol + pad])


@cuda.jit
def back_projection_gpu(recon, sinogram, angles, xlist, center, sino_height,
                        sino_width):
    """
    Implement the back-projection algorithm using GPU.

    Parameters:
    -----------
    recon : array_like
        Square array of zeros. Reconstruction image.
    sinogram : array_like
        2D array. (Filtered) sinogram image.
    angles : array_like
        1D array. Angles (radian) corresponding to the sinogram.
    xlist : array_like
        1D array. Distances of the integration lines to the image center.
    center : float
        Center of rotation.
    sino_height : int
        Height of the sinogram image.
    sino_width : int
        Width of the sinogram image.

    Returns
    -------
    recon : array_like
        Note that this is the GPU kernel function, i.e. no need of "return".
    """
    (x_index, y_index) = cuda.grid(2)
    icenter = math.ceil((sino_width - 1.0) / 2.0)
    x_cor = (x_index - icenter)
    y_cor = (y_index - icenter)
    x_min = max(-icenter, -center)
    x_max = min(sino_width - icenter - 1, sino_width - center - 1)
    if (x_index < sino_width) and (y_index < sino_width):
        num = 0.0
        for i in range(sino_height):
            theta = - angles[i]
            x_pos = x_cor * math.cos(theta) + y_cor * math.sin(theta)
            if (x_pos > x_min) and (x_pos < x_max):
                fpos = x_pos + center
                dpos = int(math.floor(fpos))
                upos = int(math.ceil(fpos))
                if upos != dpos:
                    xd = xlist[dpos]
                    xu = xlist[upos]
                    yd = sinogram[i, dpos]
                    yu = sinogram[i, upos]
                    val = yd + (yu - yd) * ((x_pos - xd) / (xu - xd))
                else:
                    val = sinogram[i, dpos]
                num += val
        recon[y_index, x_index] = num


@jit(nopython=True, parallel=True, cache=True)
def back_projection_cpu(sinogram, angles, xlist, center):
    """
    Implement the back-projection algorithm using CPU.

    Parameters:
    -----------
    sinogram : array_like
        2D array. (Filtered) sinogram image.
    angles : array_like
        1D array. Angles (radian) corresponding to the sinogram.
    xlist : array_like
        1D array. Distances of the integration lines to the image center.
    center : float
        Center of rotation.

    Returns
    -------
    recon : array_like
        Square array. Reconstructed image.
    """
    (sino_height, sino_width) = sinogram.shape
    icenter = np.ceil((sino_width - 1.0) / 2.0)
    x_min = max(-icenter, -center)
    x_max = min(sino_width - icenter - 1, sino_width - center - 1)
    recon = np.zeros((sino_width, sino_width), dtype=np.float32)
    for i in range(sino_height):
        theta = - angles[i]
        cos_theta = math.cos(theta)
        sin_theta = math.sin(theta)
        for y_index in range(sino_width):
            y_cor = y_index - icenter
            for x_index in range(sino_width):
                x_pos = (x_index - icenter) * cos_theta + y_cor * sin_theta
                if (x_pos > x_min) and (x_pos < x_max):
                    fpos = x_pos + center
                    dpos = np.int32(np.floor(fpos))
                    upos = np.int32(np.ceil(fpos))
                    if upos != dpos:
                        xd = xlist[dpos]
                        xu = xlist[upos]
                        yd = sinogram[i, dpos]
                        yu = sinogram[i, upos]
                        val = yd + (yu - yd) * ((x_pos - xd) / (xu - xd))
                    else:
                        val = sinogram[i, dpos]
                    recon[y_index, x_index] += val
    return recon


def fbp_reconstruction(sinogram, center, angles=None, ratio=1.0, ramp_win=None,
                       filter_name="hann", pad=None, pad_mode="edge",
                       apply_log=True, gpu=True):
    """
    Apply the FBP (filtered back-projection) reconstruction method to a
    sinogram image.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    center : float
        Center of rotation.
    angles : array_like, optional
        1D array. List of angles (in radian) corresponding to the sinogram.
    ratio : float, optional
        To apply a circle mask to the reconstructed image.
    ramp_win : complex ndarray, optional
        Ramp window in the Fourier space. It will be generated if None is given.
    filter_name : {None, "hann", "bartlett", "blackman", "hamming", "nuttall",\\
                  "parzen", "triang"}
        Apply a smoothing filter.
    pad : int, optional
        To apply padding before the FFT. The value is set to 10% of the image
        width if None is given.
    pad_mode : str, optional
        Padding method. Full list can be found at numpy.pad documentation.
    apply_log : bool, optional
        Apply the logarithm function to the sinogram before reconstruction.
    gpu : bool, optional
        Use GPU for computing if True.

    Returns
    -------
    array_like
        Square array. Reconstructed image.
    """
    if gpu is True:
        if cuda.is_available() is False:
            print("!!! No Nvidia GPU found !!! Run with CPU instead !!!")
            gpu = False
    if apply_log is True:
        sinogram = -np.log(sinogram)
    (nrow, ncol) = sinogram.shape
    if angles is None:
        angles = np.linspace(0.0, 180.0, nrow) * np.pi / 180.0
    else:
        num_pro = len(angles)
        if num_pro != nrow:
            msg = "!!!Number of angles is not the same as the row number of " \
                  "the sinogram!!!"
            raise ValueError(msg)
    xlist = np.float32(np.arange(0.0, ncol) - center)
    sino_filtered = apply_ramp_filter(sinogram, ramp_win, filter_name, pad,
                                      pad_mode)
    recon = np.zeros((ncol, ncol), dtype=np.float32)
    if gpu is True:
        fbp_block = (16, 16)
        fbp_grid = (int(np.ceil(1.0 * ncol / fbp_block[0])),
                    int(np.ceil(1.0 * ncol / fbp_block[1])))
        back_projection_gpu[fbp_grid, fbp_block](recon,
                                                 np.float32(sino_filtered),
                                                 np.float32(angles), xlist,
                                                 np.float32(center),
                                                 np.int32(nrow), np.int32(ncol))
    else:
        recon = back_projection_cpu(np.float32(sino_filtered),
                                    np.float32(angles), np.float32(xlist),
                                    np.float32(center))
    if ratio is not None:
        if ratio == 0.0:
            ratio = min(center, ncol - center) / (0.5 * ncol)
        mask = util.make_circle_mask(ncol, ratio)
        recon = recon * mask
    return recon * np.pi / (nrow - 1)


def generate_mapping_coordinate(width_sino, height_sino, width_rec, height_rec):
    """
    Calculate coordinates in the sinogram space from coordinates in the
    reconstruction space (in the Fourier domain). They are used for the
    DFI (direct Fourier inversion) reconstruction method.

    Parameters
    -----------
    width_sino : int
        Width of a sinogram image.
    height_sino : int
        Height of a sinogram image.
    width_rec : int
        Width of a reconstruction image.
    height_rec : int
        Height of a reconstruction image.

    Returns
    ------
    r_mat : array_like
         2D array. Broadcast of the r-coordinates.
    theta_mat : array_like
         2D array. Broadcast of the theta-coordinates.
    """
    xcenter = (width_rec - 1.0) * 0.5
    ycenter = (height_rec - 1.0) * 0.5
    r_max = np.floor(min(xcenter, ycenter))
    x_list = (np.flipud(np.arange(width_rec)) - xcenter)
    y_list = (np.arange(height_rec) - ycenter)
    x_mat, y_mat = np.meshgrid(x_list, y_list)
    r_mat = np.float32(np.clip(np.sqrt(x_mat ** 2 + y_mat ** 2), 0, r_max))
    theta_mat = np.pi + np.arctan2(y_mat, x_mat)
    r_mat[theta_mat > np.pi] *= -1
    r_mat = np.float32(np.clip(r_mat + r_max, 0, width_sino - 1))
    theta_mat[theta_mat > np.pi] -= np.pi
    theta_mat = np.float32(theta_mat * (height_sino - 1.0) / np.pi)
    return r_mat, theta_mat


def dfi_reconstruction(sinogram, center, angles=None, ratio=1.0,
                       filter_name="hann", pad_rate=0.25, pad_mode="edge",
                       apply_log=True):
    """
    Apply the DFI (direct Fourier inversion) reconstruction method to a
    sinogram image (Ref. [1]). The method is a practical and direct
    implementation of the Fourier slice theorem (Ref. [2]).

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    center : float
        Center of rotation.
    angles : array_like
        1D array. List of angles (in radian) corresponding to the sinogram.
    ratio : float
        To apply a circle mask to the reconstructed image.
    filter_name : {None, "hann", "bartlett", "blackman", "hamming", "nuttall",\\
                  "parzen", "triang"}
        Apply a smoothing filter.
    pad_rate : float
        To apply padding before the FFT. The padding width equals to
        (pad_rate * image_width).
    pad_mode : str
        Padding method. Full list can be found at numpy.pad documentation.
    apply_log : bool
        Apply the logarithm function to the sinogram before reconstruction.

    Returns
    -------
    array_like
        Square array. Reconstructed image.

    References
    ----------
    .. [1] https://doi.org/10.1364/OE.418448
    .. [2] https://doi.org/10.1071/PH560198
    """
    if apply_log is True:
        sinogram = -np.log(sinogram)
    (nrow, ncol) = sinogram.shape
    if ncol % 2 == 0:
        sinogram = np.pad(sinogram, ((0, 0), (0, 1)), mode="edge")
    ncol1 = sinogram.shape[1]
    xshift = (ncol1 - 1) / 2.0 - center
    sinogram = shift(sinogram, (0, xshift), mode='nearest')
    if angles is not None:
        t_ang = np.sum(np.abs(np.diff(angles * 180.0 / np.pi)))
        if abs(t_ang - 360) < 10:
            nrow = nrow // 2 + 1
            sinogram = (sinogram[:nrow] + np.fliplr(sinogram[-nrow:])) / 2
        step = np.mean(np.abs(np.diff(angles)))
        b_ang = angles[0] - (angles[0] // (2 * np.pi)) * (2 * np.pi)
        sino_360 = np.vstack((sinogram[: nrow - 1], np.fliplr(sinogram)))
        sinogram = shift(sino_360, (b_ang / step, 0), mode='wrap')[:nrow]
        if angles[-1] < angles[0]:
            sinogram = np.fliplr(sinogram)
    num_pad = int(pad_rate * ncol1)
    sinogram = np.pad(sinogram, ((0, 0), (num_pad, num_pad)), mode=pad_mode)
    ncol2 = sinogram.shape[1]
    mask = util.make_circle_mask(ncol2, 1.0)
    (r_mat, theta_mat) = generate_mapping_coordinate(ncol2, nrow, ncol2, ncol2)
    sino_fft = fft.fftshift(fft.fft(fft.ifftshift(sinogram, axes=1)), axes=1)
    if filter_name is not None:
        window = make_smoothing_window(filter_name, ncol2)
        sino_fft = sino_fft * np.tile(window, (nrow, 1))
    mat_real = np.real(sino_fft)
    mat_imag = np.imag(sino_fft)
    reg_real = util.mapping(mat_real, r_mat, theta_mat, order=5,
                            mode="reflect") * mask
    reg_imag = util.mapping(mat_imag, r_mat, theta_mat, order=5,
                            mode="reflect") * mask
    recon = np.real(
        fft.fftshift(fft.ifft2(fft.ifftshift(reg_real + 1j * reg_imag))))[
            num_pad:ncol + num_pad, num_pad:ncol + num_pad]
    if ratio is not None:
        if ratio == 0.0:
            ratio = min(center, ncol - center) / (0.5 * ncol)
        mask = util.make_circle_mask(ncol, ratio)
        recon = recon * mask
    return recon


def gridrec_reconstruction(sinogram, center, angles=None, ratio=1.0,
                           filter_name="shepp", apply_log=True, pad=True,
                           ncore=1):
    """
    Wrapper of the gridrec method implemented in the tomopy package:
    https://tomopy.readthedocs.io/en/latest/api/tomopy.recon.algorithm.html
    Users must install Tomopy before using this function.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    center : float
        Center of rotation.
    angles : array_like
        1D array. List of angles (radian) corresponding to the sinogram.
    ratio : float
        To apply a circle mask to the reconstructed image.
    filter_name : str
        Apply a smoothing filter. Full list is at:
        https://github.com/tomopy/tomopy/blob/master/source/tomopy/recon/algorithm.py
    apply_log : bool
        Apply the logarithm function to the sinogram before reconstruction.
    pad : bool
        Apply edge padding to the nearest power of 2.

    Returns
    -------
    array_like
        Square array.
    """
    try:
        import tomopy
    except ImportError:
        print("!!!!!! Error !!!!!!!")
        print("You must install Tomopy before using this function!")
        raise
    pad_left = 0
    ncol = sinogram.shape[-1]
    if isinstance(pad, bool):
        if pad is True:
            ncol_pad = int(2 ** np.ceil(np.log2(1.0 * ncol)))
            pad_left = (ncol_pad - ncol) // 2
            pad_right = ncol_pad - ncol - pad_left
            sinogram = np.pad(sinogram, ((0, 0), (pad_left, pad_right)),
                              mode='edge')
    else:
        pad_left = pad
        sinogram = np.pad(sinogram, ((0, 0), (pad, pad)),mode='edge')
    if apply_log is True:
        sinogram = -np.log(sinogram)
    if filter_name is None:
        filter_name = "shepp"
    if angles is None:
        angles = np.linspace(0.0, 180.0, sinogram.shape[0]) * np.pi / 180.0
    recon = tomopy.recon(np.expand_dims(sinogram, 1), angles,
                         center=center + pad_left, algorithm='gridrec',
                         filter_name=filter_name, ncore=ncore)[0]
    recon = recon[pad_left: pad_left + ncol, pad_left: pad_left + ncol]
    if ratio is not None:
        if ratio == 0.0:
            ratio = min(center, ncol - center) / (0.5 * ncol)
        mask = util.make_circle_mask(ncol, ratio)
        recon = recon * mask
    return recon


def astra_reconstruction(sinogram, center, angles=None, ratio=1.0,
                         method="FBP_CUDA", num_iter=1, filter_name="hann",
                         pad=None, apply_log=True):
    """
    Wrapper of reconstruction methods implemented in the astra toolbox package.
    https://www.astra-toolbox.com/docs/algs/index.html
    Users must install Astra Toolbox before using this function.

    Parameters
    ----------
    sinogram : array_like
        2D array. Sinogram image.
    center : float
        Center of rotation.
    angles : array_like
        1D array. List of angles (radian) corresponding to the sinogram.
    ratio : float
        To apply a circle mask to the reconstructed image.
    method : str
        Reconstruction algorithms. for CPU: 'FBP', 'SIRT', 'SART', 'ART',
        'CGLS'. for GPU: 'FBP_CUDA', 'SIRT_CUDA', 'SART_CUDA', 'CGLS_CUDA'.
    num_iter : int
        Number of iterations if using iteration methods.
    filter_name : str
        Apply filter if using FBP method. Options: 'hamming', 'hann',
        'lanczos', 'kaiser', 'parzen',...
    pad : int
        Padding to reduce the side effect of FFT.
    apply_log : bool
        Apply the logarithm function to the sinogram before reconstruction.

    Returns
    -------
    array_like
        Square array.
    """
    try:
        import astra
    except ImportError:
        print("!!!!!! Error !!!!!!!")
        print("You must install Astra Toolbox before using this function!")
        raise
    if apply_log is True:
        sinogram = -np.log(sinogram)
    if pad is None:
        pad = int(0.1 * sinogram.shape[1])
    sinogram = np.pad(sinogram, ((0, 0), (pad, pad)), mode='edge')
    (nrow, ncol) = sinogram.shape
    if angles is None:
        angles = np.linspace(0.0, 180.0, nrow) * np.pi / 180.0
    proj_geom = astra.create_proj_geom('parallel', 1, ncol, angles)
    vol_geom = astra.create_vol_geom(ncol, ncol)
    cen_col = (ncol - 1.0) / 2.0
    sinogram = shift(sinogram, (0, cen_col - (center + pad)), mode='nearest')
    sino_id = astra.data2d.create('-sino', proj_geom, sinogram)
    rec_id = astra.data2d.create('-vol', vol_geom)
    if "CUDA" not in method:
        proj_id = astra.create_projector('line', proj_geom, vol_geom)
    cfg = astra.astra_dict(method)
    cfg['ProjectionDataId'] = sino_id
    cfg['ReconstructionDataId'] = rec_id
    if "CUDA" not in method:
        cfg['ProjectorId'] = proj_id
    if (method == "FBP_CUDA") or (method == "FBP"):
        cfg["FilterType"] = filter_name
    alg_id = astra.algorithm.create(cfg)
    astra.algorithm.run(alg_id, num_iter)
    recon = astra.data2d.get(rec_id)
    astra.algorithm.delete(alg_id)
    astra.data2d.delete(sino_id)
    astra.data2d.delete(rec_id)
    recon = recon[pad:ncol - pad, pad:ncol - pad]
    if ratio is not None:
        ncol0 = ncol - 2 * pad
        if ratio == 0.0:
            ratio = min(center, ncol0 - center) / (0.5 * ncol0)
        mask = util.make_circle_mask(ncol0, ratio)
        recon = recon * mask
    return recon



# ============================================================================
# ============================================================================
# Copyright (c) 2021 Nghia T. Vo. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
# Author: Nghia T. Vo
# E-mail: algotomography@gmail.com
# Description: Python codes for loading and saving data.
# Contributors:
# ============================================================================

"""
Module for I/O tasks:
- Load data from an image file (tif, png, jpeg) or a hdf/nxs file.
- Get dataset information in a hdf/nxs file.
- Search for datasets in a hdf/nxs file.
- Save a 2D array as a tif image or 2D, 3D array to a hdf/nxs file.
- Search file names, make a file/folder name.
- Load distortion coefficients from a txt file.
"""

import os
import glob
import h5py
import numpy as np
from PIL import Image
from collections import OrderedDict


def load_image(file_path):
    """
    Load data from an image.

    Parameters
    ----------
    file_path : str
        Path to the file.

    Returns
    -------
    float
        2D array.
    """
    if "\\" in file_path:
        raise ValueError(
            "Please use the forward slash in the file path")
    try:
        mat = np.asarray(Image.open(file_path), dtype=np.float32)
    except IOError:
        print("No such file or directory: {}".format(file_path))
        raise
    if len(mat.shape) > 2:
        axis_m = np.argmin(mat.shape)
        mat = np.mean(mat, axis=axis_m)
    return mat


def get_hdf_information(file_path):
    """
    Get information of datasets in a hdf/nxs file.

    Parameters
    ----------
    file_path : str
        Path to the file.

    Returns
    -------
    list_key : str
        Keys to the datasets.
    list_shape : tuple of int
        Shapes of the datasets.
    list_type : str
        Types of the datasets.
    """
    ifile = h5py.File(file_path, 'r')
    keys = []
    ifile.visit(keys.append)
    list_key = []
    list_shape = []
    list_type = []
    for key in keys:
        data = ifile[key]
        if isinstance(data, h5py.Group):
            for key2, _ in list(data.items()):
                list_key.append(key + "/" + key2)
        else:
            list_key.append(data.name)
    for i, key in enumerate(list_key):
        data = ifile[list_key[i]]
        try:
            shape = data.shape
        except AttributeError:
            shape = "None"
        try:
            dtype = data.dtype
        except AttributeError:
            dtype = "None"
        if isinstance(data, list):
            if len(data) == 1:
                if not isinstance(data, np.ndarray):
                    dtype = str(list(data)[0])
                    dtype = dtype.replace("b'", "'")
        list_shape.append(shape)
        list_type.append(dtype)
    ifile.close()
    return list_key, list_shape, list_type


def find_hdf_key(file_path, pattern):
    """
    Find datasets matching the pattern in a hdf/nxs file.

    Parameters
    ----------
    file_path : str
        Path to the file.
    pattern : str
        Pattern to find the full names of the datasets.

    Returns
    -------
    list_key : str
        Keys to the datasets.
    list_shape : tuple of int
        Shapes of the datasets.
    list_type : str
        Types of the datasets.
    """
    ifile = h5py.File(file_path, 'r')
    list_key = []
    keys = []
    ifile.visit(keys.append)
    for key in keys:
        data = ifile[key]
        if isinstance(data, h5py.Group):
            for key2, _ in list(data.items()):
                list_key.append(data.name + "/" + key2)
        else:
            list_key.append(data.name)
    list_dkey = []
    list_dshape = []
    list_dtype = []
    for _, key in enumerate(list_key):
        if pattern in key:
            list_dkey.append(key)
            data = ifile[key]
            try:
                shape = data.shape
            except AttributeError:
                shape = "None"
            list_dshape.append(shape)
            try:
                dtype = data.dtype
            except AttributeError:
                dtype = "None"
            list_dtype.append(dtype)
            if isinstance(data, list):
                if len(data) == 1:
                    dtype = str(list(data)[0])
                    dtype = dtype.replace("b'", "'")
    ifile.close()
    return list_dkey, list_dshape, list_dtype


def load_hdf(file_path, key_path):
    """
    Load a hdf/nexus dataset as an object.

    Parameters
    ----------
    file_path : str
        Path to the file.
    key_path : str
        Key path to the dataset.

    Returns
    -------
    object
        hdf/nxs object.
    """
    try:
        ifile = h5py.File(file_path, 'r')
    except IOError:
        print("Couldn't open file: {}".format(file_path))
        raise
    check = key_path in ifile
    if not check:
        print("Couldn't open object with the key path: {}".format(key_path))
        raise ValueError("!!! Wrong key !!!")
    return ifile[key_path]


def make_folder(file_path):
    """
    Create a folder if not exist.

    Parameters
    ----------
    file_path : str
    """
    file_base = os.path.dirname(file_path)
    if not os.path.exists(file_base):
        try:
            os.makedirs(file_base)
        except OSError:
            raise ValueError("Can't create the folder: {}".format(file_path))


def make_file_name(file_path):
    """
    Create a new file name to avoid overwriting.

    Parameters
    ----------
    file_path : str

    Returns
    -------
    str
        Updated file path.
    """
    file_base, file_ext = os.path.splitext(file_path)
    if os.path.isfile(file_path):
        nfile = 0
        check = True
        while check:
            name_add = '0000' + str(nfile)
            file_path = file_base + "_" + name_add[-4:] + file_ext
            if os.path.isfile(file_path):
                nfile = nfile + 1
            else:
                check = False
    return file_path


def make_folder_name(folder_path, name_prefix="Output"):
    """
    Create a new folder name to avoid overwriting.
    E.g: Output_00001, Output_00002...

    Parameters
    ----------
    folder_path : str
        Path to the parent folder.
    name_prefix : str
        Name prefix

    Returns
    -------
    str
        Name of the folder.
    """
    zero_prefix = 5
    scan_name_prefix = name_prefix + "_"
    num_folder_exist = len(
        glob.glob(folder_path + "/" + scan_name_prefix + "*"))
    num_folder_new = num_folder_exist + 1
    name_tmp = "00000" + str(num_folder_new)
    scan_name = scan_name_prefix + name_tmp[-zero_prefix:]
    while os.path.isdir(folder_path + "/" + scan_name):
        num_folder_new = num_folder_new + 1
        name_tmp = "00000" + str(num_folder_new)
        scan_name = scan_name_prefix + name_tmp[-zero_prefix:]
    return scan_name


def find_file(path):
    """
    Search file

    Parameters
    ----------
    path : str
        Path and pattern to find files.

    Returns
    -------
    str or list of str
        List of files.
    """
    file_path = glob.glob(path)
    if len(file_path) == 0:
        raise ValueError("!!! No files found in: {}".format(path))
    for i in range(len(file_path)):
        file_path[i] = file_path[i].replace("\\", "/")
    return sorted(file_path)


def save_image(file_path, mat, overwrite=True):
    """
    Save a 2D array to an image.

    Parameters
    ----------
    file_path : str
        Path to the file.
    mat : int or float
        2D array.
    overwrite : bool
        Overwrite an existing file if True.

    Returns
    -------
    str
        Updated file path.
    """
    if "\\" in file_path:
        raise ValueError(
            "Please use the forward slash in the file path")
    _, file_ext = os.path.splitext(file_path)
    if not ((file_ext == ".tif") or (file_ext == ".tiff")):
        mat = np.uint8(255 * (mat - np.min(mat)) / (np.max(mat) - np.min(mat)))
    make_folder(file_path)
    if not overwrite:
        file_path = make_file_name(file_path)
    image = Image.fromarray(mat)
    try:
        image.save(file_path)
    except IOError:
        print("Couldn't write to file {}".format(file_path))
        raise
    return file_path


def open_hdf_stream(file_path, data_shape, key_path='entry/data',
                    data_type='float32', overwrite=True, **options):
    """
    Write an array to a hdf/nxs file with options to add metadata.

    Parameters
    ----------
    file_path : str
        Path to the file.
    data_shape : tuple of int
        Shape of the data.
    key_path : str
        Key path to the dataset.
    data_type: str
        Type of data.
    overwrite : bool
        Overwrite the existing file if True.
    options : dict, optional
        Add metadata. E.g. options={"entry/angles": angles, "entry/energy": 53}.

    Returns
    -------
    object
        hdf object.
    """
    file_base, file_ext = os.path.splitext(file_path)
    if not (file_ext == '.hdf' or file_ext == '.h5' or file_ext == ".nxs"):
        file_ext = '.hdf'
    file_path = file_base + file_ext
    make_folder(file_path)
    if not overwrite:
        file_path = make_file_name(file_path)
    try:
        ofile = h5py.File(file_path, 'w')
    except IOError:
        print("Couldn't write to file: {}".format(file_path))
        raise
    if len(options) != 0:
        for opt_name in options:
            opts = options[opt_name]
            for key in opts:
                if key_path in key:
                    msg = "!!! Selected key path, '{0}', can not be a child "\
                          "key-path of '{1}' !!!\n!!! Change to make sure they"\
                          " are at the same level !!!".format(key, key_path)
                    raise ValueError(msg)
                ofile.create_dataset(key, data=opts[key])
    data_out = ofile.create_dataset(key_path, data_shape, dtype=data_type)
    return data_out


def load_distortion_coefficient(file_path):
    """
    Load distortion coefficients from a text file.

    Parameters
    ----------
    file_path : str
        Path to the file

    Returns
    -------
    tuple of float and list
        Tuple of (xcenter, ycenter, list_fact).
    """
    if "\\" in file_path:
        raise ValueError(
            "Please use the forward slash in the file path")
    with open(file_path, 'r') as f:
        x = f.read().splitlines()
        list_data = []
        for i in x:
            list_data.append(float(i.split()[-1]))
    xcenter = list_data[0]
    ycenter = list_data[1]
    list_fact = list_data[2:]
    return xcenter, ycenter, list_fact


def save_distortion_coefficient(file_path, xcenter, ycenter, list_fact,
                                overwrite=True):
    """
    Write distortion coefficients to a text file.

    Parameters
    ----------
    file_path : str
        Path to the file.
    xcenter : float
        Center of distortion in x-direction.
    ycenter : float
        Center of distortion in y-direction.
    list_fact : float
        1D array. Coefficients of the polynomial fit.
    overwrite : bool
        Overwrite an existing file if True.

    Returns
    -------
    str
        Updated file path.
    """
    file_base, file_ext = os.path.splitext(file_path)
    if not ((file_ext == '.txt') or (file_ext == '.dat')):
        file_ext = '.txt'
    file_path = file_base + file_ext
    make_folder(file_path)
    if not overwrite:
        file_path = make_file_name(file_path)
    metadata = OrderedDict()
    metadata['xcenter'] = xcenter
    metadata['ycenter'] = ycenter
    for i, fact in enumerate(list_fact):
        kname = 'factor' + str(i)
        metadata[kname] = fact
    with open(file_path, "w") as f:
        for line in metadata:
            f.write(str(line) + " = " + str(metadata[line]))
            f.write('\n')
    return file_path
